""" DIRAC FileCatalog mix-in class to manage directory metadata
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access

__RCSID__ = "$Id$"

import six
import os
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.Time import queryTime


class DirectoryMetadata(object):
    def __init__(self, database=None):

        self.db = database

    def setDatabase(self, database):
        self.db = database

    ##############################################################################
    #
    #  Manage Metadata fields
    #

    def addMetadataField(self, pName, pType, credDict):
        """Add a new metadata parameter to the Metadata Database.

        :param str pName: parameter name
        :param str pType: parameter type in the MySQL notation

        :return: S_OK/S_ERROR, Value - comment on a positive result
        """

        result = self.db.fmeta.getFileMetadataFields(credDict)
        if not result["OK"]:
            return result
        if pName in result["Value"]:
            return S_ERROR("The metadata %s is already defined for Files" % pName)

        result = self._getMetadataFields(credDict)
        if not result["OK"]:
            return result
        if pName in result["Value"]:
            if pType.lower() == result["Value"][pName].lower():
                return S_OK("Already exists")
            return S_ERROR(
                "Attempt to add an existing metadata with different type: %s/%s" % (pType, result["Value"][pName])
            )

        valueType = pType
        if pType.lower()[:3] == "int":
            valueType = "INT"
        elif pType.lower() == "string":
            valueType = "VARCHAR(128)"
        elif pType.lower() == "float":
            valueType = "FLOAT"
        elif pType.lower() == "date":
            valueType = "DATETIME"
        elif pType == "MetaSet":
            valueType = "VARCHAR(64)"

        req = "CREATE TABLE FC_Meta_%s ( DirID INTEGER NOT NULL, Value %s, PRIMARY KEY (DirID), INDEX (Value) )" % (
            pName,
            valueType,
        )
        result = self.db._query(req)
        if not result["OK"]:
            return result

        result = self.db.insertFields("FC_MetaFields", ["MetaName", "MetaType"], [pName, pType])
        if not result["OK"]:
            return result

        metadataID = result["lastRowId"]
        result = self.__transformMetaParameterToData(pName)
        if not result["OK"]:
            return result

        return S_OK("Added new metadata: %d" % metadataID)

    def deleteMetadataField(self, pName, credDict):
        """Remove metadata field

        :param str pName: meta parameter name
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR
        """

        req = "DROP TABLE FC_Meta_%s" % pName
        result = self.db._update(req)
        error = ""
        if not result["OK"]:
            error = result["Message"]
        req = "DELETE FROM FC_MetaFields WHERE MetaName='%s'" % pName
        result = self.db._update(req)
        if not result["OK"]:
            if error:
                result["Message"] = error + "; " + result["Message"]
        return result

    def getMetadataFields(self, credDict):
        """Get all the defined metadata fields

        :param dict credDict: client credential dictionary
        :return: S_OK/S_ERROR, Value is the metadata:metadata type dictionary
        """

        return self._getMetadataFields(credDict)

    def _getMetadataFields(self, credDict):
        """Get all the defined metadata fields as they are defined in the database

        :param dict credDict: client credential dictionary
        :return: S_OK/S_ERROR, Value is the metadata:metadata type dictionary
        """

        req = "SELECT MetaName,MetaType FROM FC_MetaFields"
        result = self.db._query(req)
        if not result["OK"]:
            return result

        metaDict = {}
        for row in result["Value"]:
            metaDict[row[0]] = row[1]

        return S_OK(metaDict)

    def addMetadataSet(self, metaSetName, metaSetDict, credDict):
        """Add a new metadata set with the contents from metaSetDict

        :param str metaSetName: metaSet name
        :param dict metaSetDict: contents of the meta set definition
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR
        """
        result = self._getMetadataFields(credDict)
        if not result["OK"]:
            return result
        metaTypeDict = result["Value"]
        # Check the sanity of the metadata set contents
        for key in metaSetDict:
            if key not in metaTypeDict:
                return S_ERROR("Unknown key %s" % key)

        result = self.db.insertFields("FC_MetaSetNames", ["MetaSetName"], [metaSetName])
        if not result["OK"]:
            return result

        metaSetID = result["lastRowId"]

        req = "INSERT INTO FC_MetaSets (MetaSetID,MetaKey,MetaValue) VALUES %s"
        vList = []
        for key, value in metaSetDict.items():
            vList.append("(%d,'%s','%s')" % (metaSetID, key, str(value)))
        vString = ",".join(vList)
        result = self.db._update(req % vString)
        return result

    def getMetadataSet(self, metaSetName, expandFlag, credDict):
        """Get fully expanded contents of the metadata set

        :param str metaSetName: metaSet name
        :param bool expandFlag: flag to whether to expand the metaset recursively
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR, Value dictionary of the meta set definition contents
        """
        result = self._getMetadataFields(credDict)
        if not result["OK"]:
            return result
        metaTypeDict = result["Value"]

        req = "SELECT S.MetaKey,S.MetaValue FROM FC_MetaSets as S, FC_MetaSetNames as N "
        req += "WHERE N.MetaSetName='%s' AND N.MetaSetID=S.MetaSetID" % metaSetName
        result = self.db._query(req)
        if not result["OK"]:
            return result

        if not result["Value"]:
            return S_OK({})

        resultDict = {}
        for key, value in result["Value"]:
            if key not in metaTypeDict:
                return S_ERROR("Unknown key %s" % key)
            if expandFlag:
                if metaTypeDict[key] == "MetaSet":
                    result = self.getMetadataSet(value, expandFlag, credDict)
                    if not result["OK"]:
                        return result
                    resultDict.update(result["Value"])
                else:
                    resultDict[key] = value
            else:
                resultDict[key] = value
        return S_OK(resultDict)

    #############################################################################################
    #
    # Set and get directory metadata
    #
    #############################################################################################

    def setMetadata(self, dPath, metaDict, credDict):
        """Set the value of a given metadata field for the the given directory path

        :param str dPath: directory path
        :param dict metaDict: dictionary with metadata
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR
        """
        result = self._getMetadataFields(credDict)
        if not result["OK"]:
            return result
        metaFields = result["Value"]

        result = self.db.dtree.findDir(dPath)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Path not found: %s" % dPath)
        dirID = result["Value"]

        dirmeta = self.getDirectoryMetadata(dPath, credDict, ownData=False)
        if not dirmeta["OK"]:
            return dirmeta

        for metaName, metaValue in metaDict.items():
            if metaName not in metaFields:
                result = self.setMetaParameter(dPath, metaName, metaValue, credDict)
                if not result["OK"]:
                    return result
                continue
            # Check that the metadata is not defined for the parent directories
            if metaName in dirmeta["Value"]:
                return S_ERROR("Metadata conflict detected for %s for directory %s" % (metaName, dPath))
            result = self.db.insertFields("FC_Meta_%s" % metaName, ["DirID", "Value"], [dirID, metaValue])
            if not result["OK"]:
                if result["Message"].find("Duplicate") != -1:
                    req = "UPDATE FC_Meta_%s SET Value='%s' WHERE DirID=%d" % (metaName, metaValue, dirID)
                    result = self.db._update(req)
                    if not result["OK"]:
                        return result
                else:
                    return result

        return S_OK()

    def removeMetadata(self, dPath, metaData, credDict):
        """Remove the specified metadata for the given directory

        :param str dPath: directory path
        :param dict metaData: metadata dictionary
        :param dict credDict: client credential dictionary
        :return: standard Dirac result object
        """

        result = self._getMetadataFields(credDict)
        if not result["OK"]:
            return result
        metaFields = result["Value"]

        result = self.db.dtree.findDir(dPath)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Path not found: %s" % dPath)
        dirID = result["Value"]

        failedMeta = {}
        for meta in metaData:
            if meta in metaFields:
                # Indexed meta case
                req = "DELETE FROM FC_Meta_%s WHERE DirID=%d" % (meta, dirID)
                result = self.db._update(req)
                if not result["OK"]:
                    failedMeta[meta] = result["Value"]
            else:
                # Meta parameter case
                req = "DELETE FROM FC_DirMeta WHERE MetaKey='%s' AND DirID=%d" % (meta, dirID)
                result = self.db._update(req)
                if not result["OK"]:
                    failedMeta[meta] = result["Value"]

        if failedMeta:
            metaExample = list(failedMeta)[0]
            result = S_ERROR("Failed to remove %d metadata, e.g. %s" % (len(failedMeta), failedMeta[metaExample]))
            result["FailedMetadata"] = failedMeta
        else:
            return S_OK()

    def setMetaParameter(self, dPath, metaName, metaValue, credDict):
        """Set an meta parameter - metadata which is not used in the the data
        search operations

        :param str dPath: directory name
        :param str metaName: meta parameter name
        :param str metaValue: meta parameter value
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR
        """
        result = self.db.dtree.findDir(dPath)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Path not found: %s" % dPath)
        dirID = result["Value"]

        result = self.db.insertFields(
            "FC_DirMeta", ["DirID", "MetaKey", "MetaValue"], [dirID, metaName, str(metaValue)]
        )
        return result

    def getDirectoryMetaParameters(self, dpath, credDict, inherited=True):
        """Get meta parameters for the given directory

        :param str dPath: directory name
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR, Value dictionary of meta parameters
        """
        if inherited:
            result = self.db.dtree.getPathIDs(dpath)
            if not result["OK"]:
                return result
            pathIDs = result["Value"]
            dirID = pathIDs[-1]
        else:
            result = self.db.dtree.findDir(dpath)
            if not result["OK"]:
                return result
            if not result["Value"]:
                return S_ERROR("Path not found: %s" % dpath)
            dirID = result["Value"]
            pathIDs = [dirID]

        if len(pathIDs) > 1:
            pathString = ",".join([str(x) for x in pathIDs])
            req = "SELECT DirID,MetaKey,MetaValue from FC_DirMeta where DirID in (%s)" % pathString
        else:
            req = "SELECT DirID,MetaKey,MetaValue from FC_DirMeta where DirID=%d " % dirID
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK({})
        metaDict = {}
        for _dID, key, value in result["Value"]:
            if key in metaDict:
                if isinstance(metaDict[key], list):
                    metaDict[key].append(value)
                else:
                    metaDict[key] = [metaDict[key]].append(value)
            else:
                metaDict[key] = value

        return S_OK(metaDict)

    def getDirectoryMetadata(self, path, credDict, inherited=True, ownData=True):
        """Get metadata for the given directory aggregating metadata for the directory itself
        and for all the parent directories if inherited flag is True. Get also the non-indexed
        metadata parameters.

        :param str path: directory name
        :param dict credDict: client credential dictionary
        :param bool inherited: flag to include metadata from the parent directories
        :param bool ownData: flag to include metadata for the directory itself

        :return: S_OK/S_ERROR, Value dictionary of metadata
        """

        result = self.db.dtree.getPathIDs(path)
        if not result["OK"]:
            return result
        pathIDs = result["Value"]

        result = self._getMetadataFields(credDict)
        if not result["OK"]:
            return result
        metaFields = result["Value"]

        metaDict = {}
        metaOwnerDict = {}
        metaTypeDict = {}
        dirID = pathIDs[-1]
        if not inherited:
            pathIDs = pathIDs[-1:]
        if not ownData:
            pathIDs = pathIDs[:-1]
        pathString = ",".join([str(x) for x in pathIDs])

        for meta in metaFields:
            req = "SELECT Value,DirID FROM FC_Meta_%s WHERE DirID in (%s)" % (meta, pathString)
            result = self.db._query(req)
            if not result["OK"]:
                return result
            if len(result["Value"]) > 1:
                return S_ERROR("Metadata conflict for %s for directory %s" % (meta, path))
            if result["Value"]:
                metaDict[meta] = result["Value"][0][0]
                if int(result["Value"][0][1]) == dirID:
                    metaOwnerDict[meta] = "OwnMetadata"
                else:
                    metaOwnerDict[meta] = "ParentMetadata"
            metaTypeDict[meta] = metaFields[meta]

        # Get also non-searchable data
        result = self.getDirectoryMetaParameters(path, credDict, inherited)
        if result["OK"]:
            metaDict.update(result["Value"])
            for meta in result["Value"]:
                metaOwnerDict[meta] = "OwnParameter"

        result = S_OK(metaDict)
        result["MetadataOwner"] = metaOwnerDict
        result["MetadataType"] = metaTypeDict
        return result

    def __transformMetaParameterToData(self, metaName):
        """Relocate the meta parameters of all the directories to the corresponding
        indexed metadata table

        :param str metaName: name of the parameter to transform

        :return: S_OK/S_ERROR
        """

        req = "SELECT DirID,MetaValue from FC_DirMeta WHERE MetaKey='%s'" % metaName
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK()

        dirDict = {}
        for dirID, meta in result["Value"]:
            dirDict[dirID] = meta
        dirList = list(dirDict)

        # Exclude child directories from the list
        for dirID in dirList:
            result = self.db.dtree.getSubdirectoriesByID(dirID)
            if not result["OK"]:
                return result
            if not result["Value"]:
                continue
            childIDs = list(result["Value"])
            for childID in childIDs:
                if childID in dirList:
                    del dirList[dirList.index(childID)]

        insertValueList = []
        for dirID in dirList:
            insertValueList.append("( %d,'%s' )" % (dirID, dirDict[dirID]))

        req = "INSERT INTO FC_Meta_%s (DirID,Value) VALUES %s" % (metaName, ", ".join(insertValueList))
        result = self.db._update(req)
        if not result["OK"]:
            return result

        req = "DELETE FROM FC_DirMeta WHERE MetaKey='%s'" % metaName
        result = self.db._update(req)
        return result

    ############################################################################################
    #
    # Find directories corresponding to the metadata
    #

    def __createMetaSelection(self, value, table=""):
        """Create an SQL selection element for the given meta value

        :param dict value: dictionary with selection instructions suitable for the database search
        :param str table: table name

        :return: selection string
        """

        if isinstance(value, dict):
            selectList = []
            for operation, operand in value.items():
                if operation in [">", "<", ">=", "<="]:
                    if isinstance(operand, list):
                        return S_ERROR("Illegal query: list of values for comparison operation")
                    if isinstance(operand, six.integer_types):
                        selectList.append("%sValue%s%d" % (table, operation, operand))
                    elif isinstance(operand, float):
                        selectList.append("%sValue%s%f" % (table, operation, operand))
                    else:
                        selectList.append("%sValue%s'%s'" % (table, operation, operand))
                elif operation == "in" or operation == "=":
                    if isinstance(operand, list):
                        vString = ",".join(["'" + str(x) + "'" for x in operand])
                        selectList.append("%sValue IN (%s)" % (table, vString))
                    else:
                        selectList.append("%sValue='%s'" % (table, operand))
                elif operation == "nin" or operation == "!=":
                    if isinstance(operand, list):
                        vString = ",".join(["'" + str(x) + "'" for x in operand])
                        selectList.append("%sValue NOT IN (%s)" % (table, vString))
                    else:
                        selectList.append("%sValue!='%s'" % (table, operand))
                selectString = " AND ".join(selectList)
        elif isinstance(value, list):
            vString = ",".join(["'" + str(x) + "'" for x in value])
            selectString = "%sValue in (%s)" % (table, vString)
        else:
            if value == "Any":
                selectString = ""
            else:
                selectString = "%sValue='%s' " % (table, value)

        return S_OK(selectString)

    def __findSubdirByMeta(self, metaName, value, pathSelection="", subdirFlag=True):
        """Find directories for the given metaName datum. If the the metaName datum type is a list,
        combine values in OR. In case the metaName datum is 'Any', finds all the subdirectories
        for which the metaName datum is defined at all.

        :param str metaName: metadata name
        :param dict,list value: dictionary with selection instructions suitable for the database search
        :param str pathSelection: directory path selection string
        :param bool subdirFlag: fla to include subdirectories

        :return: S_OK/S_ERROR, Value list of found directories
        """

        result = self.__createMetaSelection(value, "M.")
        if not result["OK"]:
            return result
        selectString = result["Value"]

        req = " SELECT M.DirID FROM FC_Meta_%s AS M" % metaName
        if pathSelection:
            req += " JOIN ( %s ) AS P WHERE M.DirID=P.DirID" % pathSelection
        if selectString:
            if pathSelection:
                req += " AND %s" % selectString
            else:
                req += " WHERE %s" % selectString

        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK([])

        dirList = []
        for row in result["Value"]:
            dirID = row[0]
            dirList.append(dirID)
            # if subdirFlag:
            #  result = self.db.dtree.getSubdirectoriesByID( dirID )
            #  if not result['OK']:
            #    return result
            #  dirList += result['Value']
        if subdirFlag:
            result = self.db.dtree.getAllSubdirectoriesByID(dirList)
            if not result["OK"]:
                return result
            dirList += result["Value"]

        return S_OK(dirList)

    def __findSubdirMissingMeta(self, metaName, pathSelection):
        """Find directories not having the given meta datum defined

        :param str metaName: metadata name
        :param str pathSelection: directory path selection string

        :return: S_OK,S_ERROR , Value list of directories
        """
        result = self.__findSubdirByMeta(metaName, "Any", pathSelection)
        if not result["OK"]:
            return result
        dirList = result["Value"]
        table = self.db.dtree.getTreeTable()
        dirString = ",".join([str(x) for x in dirList])
        if dirList:
            req = "SELECT DirID FROM %s WHERE DirID NOT IN ( %s )" % (table, dirString)
        else:
            req = "SELECT DirID FROM %s" % table
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK([])

        dirList = [x[0] for x in result["Value"]]
        return S_OK(dirList)

    def __expandMetaDictionary(self, metaDict, credDict):
        """Update the dictionary with metadata query by expand metaSet type metadata

        :param dict metaDict: metaDict to be expanded
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR , Value dictionary of metadata
        """
        result = self._getMetadataFields(credDict)
        if not result["OK"]:
            return result
        metaTypeDict = result["Value"]
        resultDict = {}
        extraDict = {}
        for key, value in metaDict.items():
            if key not in metaTypeDict:
                # return S_ERROR( 'Unknown metadata field %s' % key )
                extraDict[key] = value
                continue
            keyType = metaTypeDict[key]
            if keyType != "MetaSet":
                resultDict[key] = value
            else:
                result = self.getMetadataSet(value, True, credDict)
                if not result["OK"]:
                    return result
                mDict = result["Value"]
                for mk, mv in mDict.items():
                    if mk in resultDict:
                        return S_ERROR("Contradictory query for key %s" % mk)
                    else:
                        resultDict[mk] = mv

        result = S_OK(resultDict)
        result["ExtraMetadata"] = extraDict
        return result

    def __checkDirsForMetadata(self, metaName, value, pathString):
        """Check if any of the given directories conform to the given metadata

        :param str metaName: matadata name
        :param dict,list value: dictionary with selection instructions suitable for the database search
        :param str pathString: string of comma separated directory names

        :return: S_OK/S_ERROR, Value directory ID
        """
        result = self.__createMetaSelection(value, "M.")
        if not result["OK"]:
            return result
        selectString = result["Value"]

        if selectString:
            req = "SELECT M.DirID FROM FC_Meta_%s AS M WHERE %s AND M.DirID IN (%s)" % (
                metaName,
                selectString,
                pathString,
            )
        else:
            req = "SELECT M.DirID FROM FC_Meta_%s AS M WHERE M.DirID IN (%s)" % (metaName, pathString)
        result = self.db._query(req)
        if not result["OK"]:
            return result
        elif not result["Value"]:
            return S_OK(None)
        elif len(result["Value"]) > 1:
            return S_ERROR("Conflict in the directory metadata hierarchy")
        else:
            return S_OK(result["Value"][0][0])

    @queryTime
    def findDirIDsByMetadata(self, queryDict, path, credDict):
        """Find Directories satisfying the given metadata and being subdirectories of
        the given path

        :param dict queryDict: dictionary containing query data
        :param str path: starting directory path
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR, Value list of selected directory IDs
        """

        pathDirList = []
        pathDirID = 0
        pathString = "0"
        if path != "/":
            result = self.db.dtree.getPathIDs(path)
            if not result["OK"]:
                # as result[Value] is already checked in getPathIDs
                return result
            pathIDs = result["Value"]
            pathDirID = pathIDs[-1]
            pathString = ",".join([str(x) for x in pathIDs])

        result = self.__expandMetaDictionary(queryDict, credDict)
        if not result["OK"]:
            return result
        metaDict = result["Value"]

        # Now check the meta data for the requested directory and its parents
        finalMetaDict = dict(metaDict)
        for meta in metaDict:
            result = self.__checkDirsForMetadata(meta, metaDict[meta], pathString)
            if not result["OK"]:
                return result
            elif result["Value"] is not None:
                # Some directory in the parent hierarchy is already conforming with the
                # given metadata, no need to check it further
                del finalMetaDict[meta]

        if finalMetaDict:
            pathSelection = ""
            if pathDirID:
                result = self.db.dtree.getSubdirectoriesByID(pathDirID, includeParent=True, requestString=True)
                if not result["OK"]:
                    return result
                pathSelection = result["Value"]
            dirList = []
            first = True
            for meta, value in finalMetaDict.items():
                if value == "Missing":
                    result = self.__findSubdirMissingMeta(meta, pathSelection)
                else:
                    result = self.__findSubdirByMeta(meta, value, pathSelection)
                if not result["OK"]:
                    return result
                mList = result["Value"]
                if first:
                    dirList = mList
                    first = False
                else:
                    newList = []
                    for d in dirList:
                        if d in mList:
                            newList.append(d)
                    dirList = newList
        else:
            if pathDirID:
                result = self.db.dtree.getSubdirectoriesByID(pathDirID, includeParent=True)
                if not result["OK"]:
                    return result
                pathDirList = list(result["Value"])

        finalList = []
        dirSelect = False
        if finalMetaDict:
            dirSelect = True
            finalList = dirList
            if pathDirList:
                finalList = list(set(dirList) & set(pathDirList))
        else:
            if pathDirList:
                dirSelect = True
                finalList = pathDirList
        result = S_OK(finalList)

        if finalList:
            result["Selection"] = "Done"
        elif dirSelect:
            result["Selection"] = "None"
        else:
            result["Selection"] = "All"

        return result

    @queryTime
    def findDirectoriesByMetadata(self, queryDict, path, credDict):
        """Find Directory names satisfying the given metadata and being subdirectories of
        the given path

        :param dict queryDict: dictionary containing query data
        :param str path: starting directory path
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR, Value list of selected directory paths
        """

        result = self.findDirIDsByMetadata(queryDict, path, credDict)
        if not result["OK"]:
            return result

        dirIDList = result["Value"]

        dirNameDict = {}
        if dirIDList:
            result = self.db.dtree.getDirectoryPaths(dirIDList)
            if not result["OK"]:
                return result
            dirNameDict = result["Value"]
        elif result["Selection"] == "None":
            dirNameDict = {0: "None"}
        elif result["Selection"] == "All":
            dirNameDict = {0: "All"}

        return S_OK(dirNameDict)

    def findFilesByMetadata(self, metaDict, path, credDict):
        """Find Files satisfying the given metadata

        :param dict metaDict: dictionary with the selection metadata
        :param str path: starting directory path
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR, Value list files in selected directories
        """

        result = self.findDirectoriesByMetadata(metaDict, path, credDict)
        if not result["OK"]:
            return result

        dirDict = result["Value"]
        dirList = list(dirDict)
        fileList = []
        result = self.db.dtree.getFilesInDirectory(dirList, credDict)
        if not result["OK"]:
            return result
        for _fileID, dirID, fname in result["Value"]:
            fileList.append(dirDict[dirID] + "/" + os.path.basename(fname))

        return S_OK(fileList)

    def findFileIDsByMetadata(self, metaDict, path, credDict, startItem=0, maxItems=25):
        """Find Files satisfying the given metadata

        :param dict metaDict: dictionary with the selection metadata
        :param str path: starting directory path
        :param dict credDict: client credential dictionary
        :param int startItem: offset in the file list
        :param int maxItems: max number of files to rteurn

        :return: S_OK/S_ERROR, Value list file IDs in selected directories
        """
        result = self.findDirIDsByMetadata(metaDict, path, credDict)
        if not result["OK"]:
            return result

        dirList = result["Value"]
        return self.db.dtree.getFileIDsInDirectoryWithLimits(dirList, credDict, startItem, maxItems)

    ################################################################################################
    #
    # Find metadata compatible with other metadata in order to organize dynamically updated metadata selectors

    def __findCompatibleDirectories(self, metaName, value, fromDirs):
        """Find directories compatible with the given metaName datum.
        Optionally limit the list of compatible directories to only those in the
        fromDirs list

        :param str metaName: metadata name
        :param dict,list value: dictionary with selection instructions suitable for the database search
        :param list fromDirs: list of directories to choose from

        :return: S_OK/S_ERROR, Value list of selected directories
        """

        # The directories compatible with the given metaName datum are:
        # - directory for which the datum is defined
        # - all the subdirectories of the above directory
        # - all the directories in the parent hierarchy of the above directory

        # Find directories defining the metaName datum and their subdirectories
        result = self.__findSubdirByMeta(metaName, value, subdirFlag=False)
        if not result["OK"]:
            return result
        selectedDirs = result["Value"]
        if not selectedDirs:
            return S_OK([])

        result = self.db.dtree.getAllSubdirectoriesByID(selectedDirs)
        if not result["OK"]:
            return result
        subDirs = result["Value"]

        # Find parent directories of the directories defining the metaName datum
        parentDirs = []
        for psub in selectedDirs:
            result = self.db.dtree.getPathIDsByID(psub)
            if not result["OK"]:
                return result
            parentDirs += result["Value"]

        # Constrain the output to only those that are present in the input list
        resDirs = parentDirs + subDirs + selectedDirs
        if fromDirs:
            resDirs = list(set(resDirs) & set(fromDirs))

        return S_OK(resDirs)

    def __findDistinctMetadata(self, metaList, dList):
        """Find distinct metadata values defined for the list of the input directories.
        Limit the search for only metadata in the input list

        :param list metaList: list of metadata names
        :param list dList: list of directories to limit the selection

        :return: S_OK/S_ERROR, Value dictionary of metadata
        """

        if dList:
            dString = ",".join([str(x) for x in dList])
        else:
            dString = None
        metaDict = {}
        for meta in metaList:
            req = "SELECT DISTINCT(Value) FROM FC_Meta_%s" % meta
            if dString:
                req += " WHERE DirID in (%s)" % dString
            result = self.db._query(req)
            if not result["OK"]:
                return result
            if result["Value"]:
                metaDict[meta] = []
                for row in result["Value"]:
                    metaDict[meta].append(row[0])

        return S_OK(metaDict)

    def getCompatibleMetadata(self, queryDict, path, credDict):
        """Get distinct metadata values compatible with the given already defined metadata

        :param dict queryDict: dictionary containing query data
        :param str path: starting directory path
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR, Value dictionary of metadata
        """

        pathDirID = 0
        if path != "/":
            result = self.db.dtree.findDir(path)
            if not result["OK"]:
                return result
            if not result["Value"]:
                return S_ERROR("Path not found: %s" % path)
            pathDirID = int(result["Value"])
        pathDirs = []
        if pathDirID:
            result = self.db.dtree.getSubdirectoriesByID(pathDirID, includeParent=True)
            if not result["OK"]:
                return result
            if result["Value"]:
                pathDirs = list(result["Value"])
            result = self.db.dtree.getPathIDsByID(pathDirID)
            if not result["OK"]:
                return result
            if result["Value"]:
                pathDirs += result["Value"]

        # Get the list of metadata fields to inspect
        result = self._getMetadataFields(credDict)
        if not result["OK"]:
            return result
        metaFields = result["Value"]
        comFields = list(metaFields)

        # Commented out to return compatible data also for selection metadata
        # for m in metaDict:
        #  if m in comFields:
        #    del comFields[comFields.index( m )]

        result = self.__expandMetaDictionary(queryDict, credDict)
        if not result["OK"]:
            return result
        metaDict = result["Value"]

        fromList = pathDirs
        anyMeta = True
        if metaDict:
            anyMeta = False
            for meta, value in metaDict.items():
                result = self.__findCompatibleDirectories(meta, value, fromList)
                if not result["OK"]:
                    return result
                cdirList = result["Value"]
                if cdirList:
                    fromList = cdirList
                else:
                    fromList = []
                    break

        if anyMeta or fromList:
            result = self.__findDistinctMetadata(comFields, fromList)
        else:
            result = S_OK({})
        return result

    def removeMetadataForDirectory(self, dirList, credDict):
        """Remove all the metadata for the given directory list

        :param list dirList: list of directory paths
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR, Value Successful/Failed dictionaries
        """
        if not dirList:
            return S_OK({"Successful": {}, "Failed": {}})

        failed = {}
        successful = {}
        dirs = dirList
        if not isinstance(dirList, list):
            dirs = [dirList]

        dirListString = ",".join([str(d) for d in dirs])

        # Get the list of metadata fields to inspect
        result = self._getMetadataFields(credDict)
        if not result["OK"]:
            return result
        metaFields = result["Value"]

        for meta in metaFields:
            req = "DELETE FROM FC_Meta_%s WHERE DirID in ( %s )" % (meta, dirListString)
            result = self.db._query(req)
            if not result["OK"]:
                failed[meta] = result["Message"]
            else:
                successful[meta] = "OK"

        return S_OK({"Successful": successful, "Failed": failed})



"""
AccountingCLI class implementing command line administrative interface to
DIRAC Accounting DataStore Service
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import sys

import six

from DIRAC import gLogger
from DIRAC.Core.Base.CLI import CLI, colorize
from DIRAC.AccountingSystem.Client.DataStoreClient import DataStoreClient


class AccountingCLI(CLI):
    def __init__(self):
        CLI.__init__(self)
        self.do_connect(None)

    def start(self):
        """
        Start the command loop
        """
        if not self.connected:
            gLogger.error("Client is not connected")
        try:
            self.cmdloop()
        except KeyboardInterrupt:
            gLogger.warn("Received a keyboard interrupt.")
            self.do_quit("")

    def do_connect(self, args):
        """
        Tries to connect to the server
            Usage: connect
        """
        gLogger.info("Trying to connect to server")
        self.connected = False
        self.prompt = "(%s)> " % colorize("Not connected", "red")
        acClient = DataStoreClient()
        retVal = acClient.ping()
        if retVal["OK"]:
            self.prompt = "(%s)> " % colorize("Connected", "green")
            self.connected = True

    def printComment(self, comment):
        commentList = comment.split("\n")
        for commentLine in commentList[:-1]:
            print("# %s" % commentLine.strip())

    def showTraceback(self):
        import traceback

        type, value = sys.exc_info()[:2]
        print("________________________\n")
        print("Exception", type, ":", value)
        traceback.print_tb(sys.exc_info()[2])
        print("________________________\n")

    def do_registerType(self, args):
        """
        Registers a new accounting type
          Usage : registerType <typeName>
          <DIRACRoot>/DIRAC/AccountingSystem/Client/Types/<typeName>
           should exist and inherit the base type
        """
        try:
            argList = args.split()
            if argList:
                typeName = argList[0].strip()
            else:
                gLogger.error("No type name specified")
                return
            # Try to import the type
            try:
                typeModule = __import__(
                    "DIRAC.AccountingSystem.Client.Types.%s" % typeName, globals(), locals(), typeName
                )
                typeClass = getattr(typeModule, typeName)
            except Exception as e:
                gLogger.error("Can't load type %s: %s" % (typeName, str(e)))
                return
            gLogger.info("Loaded type %s" % typeClass.__name__)
            typeDef = typeClass().getDefinition()
            acClient = DataStoreClient()
            retVal = acClient.registerType(*typeDef)
            if retVal["OK"]:
                gLogger.info("Type registered successfully")
            else:
                gLogger.error("Error: %s" % retVal["Message"])
        except Exception:
            self.showTraceback()

    def do_resetBucketLength(self, args):
        """
        Set the bucket Length. Will trigger a recalculation of buckets. Can take a while.
          Usage : resetBucketLength <typeName>
          <DIRACRoot>/DIRAC/AccountingSystem/Client/Types/<typeName>
           should exist and inherit the base type
        """
        try:
            argList = args.split()
            if argList:
                typeName = argList[0].strip()
            else:
                gLogger.error("No type name specified")
                return
            # Try to import the type
            try:
                typeModule = __import__(
                    "DIRAC.AccountingSystem.Client.Types.%s" % typeName, globals(), locals(), typeName
                )
                typeClass = getattr(typeModule, typeName)
            except Exception as e:
                gLogger.error("Can't load type %s: %s" % (typeName, str(e)))
                return
            gLogger.info("Loaded type %s" % typeClass.__name__)
            typeDef = typeClass().getDefinition()
            acClient = DataStoreClient()
            retVal = acClient.setBucketsLength(typeDef[0], typeDef[3])
            if retVal["OK"]:
                gLogger.info("Type registered successfully")
            else:
                gLogger.error("Error: %s" % retVal["Message"])
        except Exception:
            self.showTraceback()

    def do_regenerateBuckets(self, args):
        """
        Regenerate buckets for type. Can take a while.
          Usage : regenerateBuckets <typeName>
          <DIRACRoot>/DIRAC/AccountingSystem/Client/Types/<typeName>
           should exist and inherit the base type
        """
        try:
            argList = args.split()
            if argList:
                typeName = argList[0].strip()
            else:
                gLogger.error("No type name specified")
                return
            # Try to import the type
            try:
                typeModule = __import__(
                    "DIRAC.AccountingSystem.Client.Types.%s" % typeName, globals(), locals(), typeName
                )
                typeClass = getattr(typeModule, typeName)
            except Exception as e:
                gLogger.error("Can't load type %s: %s" % (typeName, str(e)))
                return
            gLogger.info("Loaded type %s" % typeClass.__name__)
            typeDef = typeClass().getDefinition()
            acClient = DataStoreClient()
            retVal = acClient.regenerateBuckets(typeDef[0])
            if retVal["OK"]:
                gLogger.info("Buckets recalculated!")
            else:
                gLogger.error("Error: %s" % retVal["Message"])
        except Exception:
            self.showTraceback()

    def do_showRegisteredTypes(self, args):
        """
        Get a list of registered types
          Usage : showRegisteredTypes
        """
        try:
            acClient = DataStoreClient()
            retVal = acClient.getRegisteredTypes()

            print(retVal)

            if not retVal["OK"]:
                gLogger.error("Error: %s" % retVal["Message"])
                return
            for typeList in retVal["Value"]:
                print(typeList[0])
                print(" Key fields:\n  %s" % "\n  ".join(typeList[1]))
                print(" Value fields:\n  %s" % "\n  ".join(typeList[2]))
        except Exception:
            self.showTraceback()

    def do_deleteType(self, args):
        """
        Delete a registered accounting type.
          Usage : deleteType <typeName>
          WARN! It will delete all data associated to that type! VERY DANGEROUS!
          If you screw it, you'll discover a new dimension of pain and doom! :)
        """
        try:
            argList = args.split()
            if argList:
                typeName = argList[0].strip()
            else:
                gLogger.error("No type name specified")
                return
            while True:
                choice = six.moves.input(
                    "Are you completely sure you want to delete type %s and all it's data? yes/no [no]: " % typeName
                )
                choice = choice.lower()
                if choice in ("yes", "y"):
                    break
                else:
                    print("Delete aborted")
                    return
            acClient = DataStoreClient()
            retVal = acClient.deleteType(typeName)
            if not retVal["OK"]:
                gLogger.error("Error: %s" % retVal["Message"])
                return
            print("Hope you meant it, because it's done")
        except Exception:
            self.showTraceback()

    def do_compactBuckets(self, args):
        """
        Compact buckets table
          Usage : compactBuckets
        """
        try:
            acClient = DataStoreClient()
            retVal = acClient.compactDB()
            if not retVal["OK"]:
                gLogger.error("Error: %s" % retVal["Message"])
                return
            gLogger.info("Done")
        except Exception:
            self.showTraceback()



""" ProxyProvider implementation for the proxy generation using local (DIRAC) CA credentials

    This class is a simple, limited CA, its main purpose is to generate a simple proxy for DIRAC users
    who do not have any certificate register on the fly.

    Required parameters in the DIRAC configuration for its implementation:

    .. literalinclude:: /dirac.cfg
      :start-after: ## DIRACCA type:
      :end-before: ##
      :dedent: 2
      :caption: /Resources/ProxyProviders section

    Also, as an additional feature, this class can read properties from a simple openssl CA configuration file.
    To do this, just set the path to an existing configuration file as a CAConfigFile parameter. In this case,
    the distinguished names order in the created proxy will be the same as in the configuration file policy block.

    The Proxy provider supports the following distinguished names
    (https://www.cryptosys.net/pki/manpki/pki_distnames.html)::

      SN(surname)
      GN(givenName)
      C(countryName)
      CN(commonName)
      L(localityName)
      Email(emailAddress)
      O(organizationName)
      OU(organizationUnitName)
      SP,ST(stateOrProvinceName)
      SERIALNUMBER(serialNumber)

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import re
import time
import random
import datetime
import collections

from M2Crypto import m2, util, X509, ASN1, EVP, RSA

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Security.X509Chain import X509Chain  # pylint: disable=import-error
from DIRAC.Resources.ProxyProvider.ProxyProvider import ProxyProvider


class DIRACCAProxyProvider(ProxyProvider):
    def __init__(self, parameters=None):
        """Constructor"""
        super(DIRACCAProxyProvider, self).__init__(parameters)
        self.log = gLogger.getSubLogger(__name__)
        # Initialize
        self.maxDict = {}
        self.minDict = {}
        self.bits = 2048
        self.algoritm = "sha256"
        self.match = []
        self.supplied = ["CN"]
        self.optional = ["C", "O", "OU", "emailAddress"]
        self.dnList = ["C", "O", "OU", "CN", "emailAddress"]
        # Distinguished names
        self.fields2nid = X509.X509_Name.nid.copy()
        self.fields2nid["DC"] = -1  # Add DN that is not liested in X509.X509_Name
        self.fields2nid["domainComponent"] = -1  # Add DN description that is not liested in X509.X509_Name
        self.fields2nid["organizationalUnitName"] = 18  # Add 'OU' description
        self.fields2nid["countryName"] = 14  # Add 'C' description
        self.fields2nid["SERIALNUMBER"] = 105  # Add 'SERIALNUMBER' distinguished name
        self.nid2fields = {}  # nid: list of distinguished names
        # Specify standart fields
        for field, nid in self.fields2nid.items():
            self.nid2fields.setdefault(nid, []).append(field)
        self.dnInfoDictCA = {}

    def setParameters(self, parameters):
        """Set new parameters

        :param dict parameters: provider parameters

        :return: S_OK()/S_ERROR()
        """
        for k, v in parameters.items():
            if not isinstance(v, list) and k in ["Match", "Supplied", "Optional", "DNOrder"] + list(self.fields2nid):
                parameters[k] = v.replace(", ", ",").split(",")
        self.parameters = parameters
        # If CA configuration file exist
        if parameters.get("CAConfigFile"):
            self.__parseCACFG()
        if "Bits" in parameters:
            self.bits = int(parameters["Bits"])
        if "Algoritm" in parameters:
            self.algoritm = parameters["Algoritm"]
        if "Match" in parameters:
            self.match = [self.fields2nid[f] for f in parameters["Match"]]
        if "Supplied" in parameters:
            self.supplied = [self.fields2nid[f] for f in parameters["Supplied"]]
        if "Optional" in parameters:
            self.optional = [self.fields2nid[f] for f in parameters["Optional"]]
        allFields = self.optional + self.supplied + self.match
        if "DNOrder" in parameters:
            self.dnList = []
            if not any([any([f in parameters["DNOrder"] for f in self.nid2fields[n]]) for n in allFields]):
                return S_ERROR("DNOrder must contain all configured fields.")
            for field in parameters["DNOrder"]:
                if self.fields2nid[field] in allFields:
                    self.dnList.append(field)

        # Set defaults for distridutes names
        self.nid2defField = {}
        for field, value in list(self.parameters.items()):
            if field in self.fields2nid and self.fields2nid[field] in allFields:
                self.parameters[self.fields2nid[field]] = value
                self.nid2defField[self.fields2nid[field]] = field

        # Read CA certificate
        chain = X509Chain()
        result = chain.loadChainFromFile(self.parameters["CertFile"])
        if result["OK"]:
            result = chain.getCredentials()
            if result["OK"]:
                result = self.__parseDN(result["Value"]["subject"])
        if not result["OK"]:
            return result
        self.dnInfoDictCA = result["Value"]
        return S_OK()

    def checkStatus(self, userDN):
        """Read ready to work status of proxy provider

        :param str userDN: user DN

        :return: S_OK()/S_ERROR()
        """
        self.log.debug("Ckecking work status of", self.parameters["ProviderName"])
        result = self.__parseDN(userDN)
        if not result["OK"]:
            return result
        dnInfoDict = result["Value"]

        try:
            userNIDs = [self.fields2nid[f.split("=")[0]] for f in userDN.lstrip("/").split("/")]
        except (ValueError, KeyError) as e:
            return S_ERROR("Unknown DN field in used DN: %s" % e)
        nidOrder = [self.fields2nid[f] for f in self.dnList]
        for index, nid in enumerate(userNIDs):
            if nid not in nidOrder:
                return S_ERROR(
                    '"%s" field not found in order.' % self.nid2defField.get(nid, min(self.nid2fields[nid], key=len))
                )
            if index > nidOrder.index(nid):
                return S_ERROR("Bad DNs order")
            for i in range(nidOrder.index(nid) - 1):
                try:
                    if userNIDs.index(nidOrder[i]) > index:
                        return S_ERROR("Bad DNs order")
                except (ValueError, KeyError):
                    continue
            for i in range(nidOrder.index(nid) + 1, len(nidOrder)):
                try:
                    if userNIDs.index(nidOrder[i]) < index:
                        return S_ERROR("Bad DNs order")
                except (ValueError, KeyError):
                    continue

        for nid in self.supplied:
            if nid not in [self.fields2nid[f] for f in dnInfoDict]:
                return S_ERROR(
                    'Current DN is invalid, "%s" field must be set.'
                    % self.nid2defField.get(nid, min(self.nid2fields[nid], key=len))
                )

        for field, values in dnInfoDict.items():
            nid = self.fields2nid[field]
            err = 'Current DN is invalid, "%s" field' % field
            if nid not in self.supplied + self.match + self.optional:
                return S_ERROR("%s is not found for current CA." % err)
            if nid in self.match and not self.dnInfoDictCA[field] == values:
                return S_ERROR("%s must be /%s=%s." % (err, field, ("/%s=" % field).joing(self.dnInfoDictCA[field])))
            if nid in self.maxDict:
                rangeMax = list(range(min(len(values), len(self.maxDict[nid]))))
                if any([True if len(values[i]) > self.maxDict[nid][i] else False for i in rangeMax]):
                    return S_ERROR("%s values must be less then %s." % (err, ", ".join(self.maxDict[nid])))
            if nid in self.minDict:
                rangeMin = list(range(min(len(values), len(self.minDict[nid]))))
                if any([True if len(values[i]) < self.minDict[nid][i] else False for i in rangeMin]):
                    return S_ERROR("%s values must be more then %s." % (err, ", ".join(self.minDict[nid])))

            result = self.__fillX509Name(field, values)
            if not result["OK"]:
                return result

        return S_OK()

    def getProxy(self, userDN):
        """Generate user proxy

        :param str userDN: user DN

        :return: S_OK(str)/S_ERROR() -- contain a proxy string
        """
        self.__X509Name = X509.X509_Name()
        result = self.checkStatus(userDN)
        if result["OK"]:
            result = self.__createCertM2Crypto()
            if result["OK"]:
                certStr, keyStr = result["Value"]

                chain = X509Chain()
                result = chain.loadChainFromString(certStr)
                if result["OK"]:
                    result = chain.loadKeyFromString(keyStr)
                    if result["OK"]:
                        result = chain.generateProxyToString(365 * 24 * 3600, rfc=True)

        return result

    def generateDN(self, **kwargs):
        """Get DN of the user certificate that will be created

        :param dict kwargs: user description dictionary with possible fields:
               - FullName or CN
               - Email or emailAddress

        :return: S_OK(str)/S_ERROR() -- contain DN
        """
        if kwargs.get("FullName"):
            kwargs["CN"] = [kwargs["FullName"]]
        if kwargs.get("Email"):
            kwargs["emailAddress"] = [kwargs["Email"]]

        self.__X509Name = X509.X509_Name()
        self.log.info("Creating distinguished names chain")

        for nid in self.supplied:
            if nid not in [self.fields2nid[f] for f in self.dnList]:
                return S_ERROR(
                    'DNs order list does not contain supplied DN "%s"'
                    % self.nid2defField.get(nid, min(self.nid2fields[nid], key=len))
                )

        for field in self.dnList:
            values = []
            nid = self.fields2nid[field]
            if nid in self.match:
                for field in self.nid2fields[nid]:
                    if field in self.dnInfoDictCA:
                        values = self.dnInfoDictCA[field]
                if not values:
                    return S_ERROR('Not found "%s" match DN in CA' % field)
            for field in self.nid2fields[nid]:
                if kwargs.get(field):
                    values = kwargs[field] if isinstance(kwargs[field], list) else [kwargs[field]]
            if not values and nid in self.supplied:
                # Search default value
                if nid not in self.nid2defField:
                    return S_ERROR('No values set for "%s" DN' % min(self.nid2fields[nid], key=len))
                values = self.parameters[nid]

            result = self.__fillX509Name(field, values)
            if not result["OK"]:
                return result

        # WARN: This logic not support list of distribtes name elements
        resDN = m2.x509_name_oneline(self.__X509Name.x509_name)  # pylint: disable=no-member

        result = self.checkStatus(resDN)
        if not result["OK"]:
            return result
        return S_OK(resDN)

    def __parseCACFG(self):
        """Parse CA configuration file"""
        block = ""
        self.cfg = {}
        self.supplied, self.optional, self.match, self.dnList = [], [], [], []
        with open(self.parameters["CAConfigFile"], "r") as caCFG:
            for line in caCFG:
                # Ignore comments
                line = re.sub(r"#.*", "", line)
                if re.findall(r"\[([A-Za-z0-9_]+)\]", line.replace(" ", "")):
                    block = "".join(re.findall(r"\[([A-Za-z0-9_]+)\]", line.replace(" ", "")))
                    if block not in self.cfg:
                        self.cfg[block] = {}
                if not block:
                    continue
                if len(re.findall("=", line)) == 1:
                    field, val = line.split("=")
                    field = field.strip()
                    variables = re.findall(r"[$]([A-Za-z0-9_]+)", val)
                    for v in variables:
                        for b in self.cfg:
                            if v in self.cfg[b]:
                                val = val.replace("$" + v, self.cfg[b][v])
                    if "default_ca" in self.cfg.get("ca", {}):
                        if "policy" in self.cfg.get(self.cfg["ca"]["default_ca"], {}):
                            if block == self.cfg[self.cfg["ca"]["default_ca"]]["policy"]:
                                self.dnList.append(field)
                    self.cfg[block][field] = val.strip()

        self.bits = int(self.cfg["req"].get("default_bits") or self.bits)
        self.algoritm = self.cfg[self.cfg["ca"]["default_ca"]].get("default_md") or self.algoritm
        if not self.parameters.get("CertFile"):
            self.parameters["CertFile"] = self.cfg[self.cfg["ca"]["default_ca"]]["certificate"]
            self.parameters["KeyFile"] = self.cfg[self.cfg["ca"]["default_ca"]]["private_key"]
        # Read distinguished names
        for k, v in self.cfg[self.cfg[self.cfg["ca"]["default_ca"]]["policy"]].items():
            nid = self.fields2nid[k]
            self.parameters[nid], self.minDict[nid], self.maxDict[nid] = [], [], []
            for k in ["%s.%s" % (i, k) for i in range(0, 5)] + [k]:
                if k + "_default" in self.cfg["req"]["distinguished_name"]:
                    self.parameters[nid].append(self.cfg["req"]["distinguished_name"][k + "_default"])
                if k + "_min" in self.cfg["req"]["distinguished_name"]:
                    self.minDict[nid].append(self.cfg["req"]["distinguished_name"][k + "_min"])
                if k + "_max" in self.cfg["req"]["distinguished_name"]:
                    self.maxDict[nid].append(self.cfg["req"]["distinguished_name"][k + "_max"])
            if v == "supplied":
                self.supplied.append(nid)
            elif v == "optional":
                self.optional.append(nid)
            elif v == "match":
                self.match.append(nid)

    def __parseDN(self, dn):
        """Return DN fields

        :param str dn: DN

        :return: list -- contain tuple with positionOfField.fieldName, fieldNID, fieldValue
        """
        dnInfoDict = collections.OrderedDict()
        for f, v in [f.split("=") for f in dn.lstrip("/").split("/")]:
            if not v:
                return S_ERROR('No value set for "%s"' % f)
            if f not in dnInfoDict:
                dnInfoDict[f] = [v]
            else:
                dnInfoDict[f].append(v)
        return S_OK(dnInfoDict)

    def __fillX509Name(self, field, values):
        """Fill x509_Name object by M2Crypto

        :param str field: DN field name
        :param list values: values of field, order important

        :return: S_OK()/S_ERROR()
        """
        for value in values:
            if (
                value
                and m2.x509_name_set_by_nid(  # pylint: disable=no-member
                    self.__X509Name.x509_name, self.fields2nid[field], value.encode()
                )
                == 0
            ):
                if (
                    not self.__X509Name.add_entry_by_txt(
                        field=field, type=ASN1.MBSTRING_ASC, entry=value, len=-1, loc=-1, set=0
                    )
                    == 1
                ):
                    return S_ERROR('Cannot set "%s" field.' % field)
        return S_OK()

    def __createCertM2Crypto(self):
        """Create new certificate for user

        :return: S_OK(tuple)/S_ERROR() -- tuple contain certificate and pulic key as strings
        """
        # Create public key
        userPubKey = EVP.PKey()
        userPubKey.assign_rsa(RSA.gen_key(self.bits, 65537, util.quiet_genparam_callback))
        # Create certificate
        userCert = X509.X509()
        userCert.set_pubkey(userPubKey)
        userCert.set_version(2)
        userCert.set_subject(self.__X509Name)
        userCert.set_serial_number(int(random.random() * 10 ** 10))
        # Add extentionals
        userCert.add_ext(X509.new_extension("basicConstraints", "CA:" + str(False).upper()))
        userCert.add_ext(X509.new_extension("extendedKeyUsage", "clientAuth", critical=1))
        # Set livetime
        validityTime = datetime.timedelta(days=400)
        notBefore = ASN1.ASN1_UTCTIME()
        notBefore.set_time(int(time.time()))
        notAfter = ASN1.ASN1_UTCTIME()
        notAfter.set_time(int(time.time()) + int(validityTime.total_seconds()))
        userCert.set_not_before(notBefore)
        userCert.set_not_after(notAfter)
        # Add subject from CA
        with open(self.parameters["CertFile"]) as cf:
            caCertStr = cf.read()
        caCert = X509.load_cert_string(caCertStr)
        userCert.set_issuer(caCert.get_subject())
        # Use CA key
        with open(self.parameters["KeyFile"], "rb") as cf:
            caKeyStr = cf.read()
        pkey = EVP.PKey()
        pkey.assign_rsa(RSA.load_key_string(caKeyStr, callback=util.no_passphrase_callback))
        # Sign
        userCert.sign(pkey, self.algoritm)

        userCertStr = userCert.as_pem()
        userPubKeyStr = userPubKey.as_pem(cipher=None, callback=util.no_passphrase_callback)
        return S_OK((userCertStr, userPubKeyStr))

    def _forceGenerateProxyForDN(self, dn, time, group=None):
        """An additional helper method for creating a proxy without any substantial validation,
        it can be used for a specific case(such as testing) where just need to generate a proxy
        with specific DN on the fly.

        :param str dn: requested proxy DN
        :param int time: expired time in a seconds
        :param str group: if need to add DIRAC group

        :return: S_OK(tuple)/S_ERROR() -- contain proxy as chain and as string
        """
        self.__X509Name = X509.X509_Name()
        result = self.__parseDN(dn)
        if not result["OK"]:
            return result
        dnInfoDict = result["Value"]

        for field, values in dnInfoDict.items():
            result = self.__fillX509Name(field, values)
            if not result["OK"]:
                return result

        result = self.__createCertM2Crypto()
        if result["OK"]:
            certStr, keyStr = result["Value"]
            chain = X509Chain()
            if chain.loadChainFromString(certStr)["OK"] and chain.loadKeyFromString(keyStr)["OK"]:
                result = chain.generateProxyToString(time, rfc=True, diracGroup=group)
        if not result["OK"]:
            return result
        chain = X509Chain()
        chain.loadProxyFromString(result["Value"])
        return S_OK((chain, result["Value"]))



""" DIRAC FileCatalog component representing a directory tree with simple nodes
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os

from DIRAC import S_OK, S_ERROR
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.DirectoryManager.DirectoryTreeBase import DirectoryTreeBase


class DirectoryNodeTree(DirectoryTreeBase):
    """Class managing Directory Tree as a self-linked structure with directory
    names stored in each node
    """

    def __init__(self, database=None):
        DirectoryTreeBase.__init__(self, database)
        self.treeTable = "FC_DirectoryTreeM"

    def findDir(self, path):
        """Find the identifier of a directory specified by its path"""
        dpath = path
        if path[0] == "/":
            dpath = path[1:]
        elements = dpath.split("/")

        req = " "
        for level in range(len(elements), 0, -1):
            if level > 1:
                req += "SELECT DirID from FC_DirectoryTreeM WHERE Level=%d AND DirName='%s' AND Parent=(" % (
                    level,
                    elements[level - 1],
                )
            else:
                req += "SELECT DirID from FC_DirectoryTreeM WHERE Level=%d AND DirName='%s'" % (
                    level,
                    elements[level - 1],
                )
        req += ")" * (len(elements) - 1)
        # print req
        result = self.db._query(req)
        # print "in findDir",result
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK(0)

        return S_OK(result["Value"][0][0])

    def makeDir(self, path):
        """Create a single directory"""
        result = self.findDir(path)
        if not result["OK"]:
            return result
        dirID = result["Value"]
        if dirID:
            return S_OK(dirID)

        dpath = path
        if path[0] == "/":
            dpath = path[1:]
        elements = dpath.split("/")
        level = len(elements)
        dirName = elements[-1]

        result = self.getParent(path)
        if not result["OK"]:
            return result
        parentDirID = result["Value"]

        names = ["DirName", "Level", "Parent"]
        values = [dirName, level, parentDirID]
        result = self.db.insertFields("FC_DirectoryTreeM", names, values)
        if not result["OK"]:
            return result
        return S_OK(result["lastRowId"])

    def existsDir(self, path):
        """Check the existence of a directory at the specified path"""
        result = self.findDir(path)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK({"Exists": False})
        else:
            return S_OK({"Exists": True, "DirID": result["Value"]})

    def getParent(self, path):
        """Get the parent ID of the given directory"""
        dpath = path
        if path[0] == "/":
            dpath = path[1:]
        elements = dpath.split("/")
        if len(elements) > 1:
            parentDir = os.path.dirname(path)
            result = self.findDir(parentDir)
            if not result["OK"]:
                return result
            parentDirID = result["Value"]
            if not parentDirID:
                return S_ERROR("No parent directory")
            return S_OK(parentDirID)
        else:
            return S_OK(0)

    def getParentID(self, dirID):
        """ """

        if dirID == 0:
            return S_ERROR("Root directory ID given")

        req = "SELECT Parent FROM FC_DirectoryTreeM WHERE DirID=%d" % dirID
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("No parent found")

        return S_OK(result["Value"][0][0])

    def getDirectoryName(self, dirID):
        """Get directory name by directory ID"""
        req = "SELECT DirName FROM FC_DirectoryTreeM WHERE DirID=%d" % int(dirID)
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Directory with id %d not found" % int(dirID))

        return S_OK(result["Value"][0][0])

    def getDirectoryPath(self, dirID):
        """Get directory path by directory ID"""

        dirPath = ""
        dID = dirID
        while True:
            result = self.getDirectoryName(dID)
            if not result["OK"]:
                return result
            dirPath = "/" + result["Value"] + dirPath
            result = self.getParentID(dID)
            if not result["OK"]:
                return result
            if result["Value"] == 0:
                break
            else:
                dID = result["Value"]

        return S_OK("/" + dirPath)

    def getPathIDs(self, path):
        """Get IDs of all the directories in the parent hierarchy"""
        result = self.findDir(path)
        if not result["OK"]:
            return result
        dID = result["Value"]
        parentIDs = []
        while True:
            result = self.getParent(dID)
            if not result["OK"]:
                return result
            dID = result["Value"]
            parentIDs.append(dID)
            if dID == 0:
                break

        parentIDs.append(0)
        parentIDs.reverse()
        return S_OK(parentIDs)

    def getChildren(self, path):
        """Get child directory IDs for the given directory"""
        if isinstance(path, str):
            result = self.findDir(path)
            if not result["OK"]:
                return result
            dirID = result["Value"]
        else:
            dirID = path

        req = "SELECT DirID FROM FC_DirectoryTreeM WHERE Parent=%d" % dirID
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK([])

        return S_OK([x[0] for x in result["Value"]])



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import random
import threading
from hashlib import md5

from DIRAC.Core.Utilities.ThreadSafe import Synchronizer
from DIRAC.Core.DISET.private.BaseClient import BaseClient
from DIRAC.Core.DISET.private.MessageBroker import getGlobalMessageBroker
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR, isReturnStructure
from DIRAC.Core.Utilities import Network, Time
from DIRAC.FrameworkSystem.Client.Logger import gLogger

gMsgSync = Synchronizer()


class MessageClient(BaseClient):
    class MSGException(Exception):
        pass

    def _initialize(self):
        self.__trid = False
        self.__transport = None
        self.__uniqueName = self.__generateUniqueClientName()
        self.__msgBroker = getGlobalMessageBroker()
        self.__callbacks = {}
        self.__connectExtraParams = {}
        self.__specialCallbacks = {"drop": [], "msg": []}
        self.__connectionLock = threading.Lock()

    def __generateUniqueClientName(self):
        hashStr = ":".join((Time.toString(), str(random.random()), Network.getFQDN(), gLogger.getName()))
        hexHash = md5(hashStr.encode()).hexdigest()
        return hexHash

    def setUniqueName(self, uniqueName):
        self.__uniqueName = uniqueName

    def __checkResult(self, result):
        if not result["OK"]:
            raise self.MSGException(result["Message"])
        return result["Value"]

    def createMessage(self, msgName):
        return self.__msgBroker.getMsgFactory().createMessage(self.getServiceName(), msgName)

    @property
    def connected(self):
        return self.__trid

    def connect(self, **extraParams):
        if extraParams:
            self.__connectExtraParams = extraParams
        # Avoid acquiring the lock if it's clear we already have a connection
        if self.__trid:
            return S_ERROR("Already connected")
        self.__connectionLock.acquire()
        try:
            if self.__trid:
                return S_ERROR("Already connected")

            trid, transport = self.__checkResult(self._connect())
            self.__checkResult(self._proposeAction(transport, ("Connection", "new")))
            self.__checkResult(transport.sendData(S_OK([self.__uniqueName, self.__connectExtraParams])))
            self.__checkResult(transport.receiveData())
            self.__checkResult(
                self.__msgBroker.addTransportId(
                    trid,
                    self._serviceName,
                    receiveMessageCallback=self.__cbRecvMsg,
                    disconnectCallback=self.__cbDisconnect,
                )
            )
            self.__trid = trid
            self.__transport = transport
        except self.MSGException as e:
            return S_ERROR(str(e))
        finally:
            self.__connectionLock.release()
        return S_OK()

    def __cbDisconnect(self, trid):
        if not self.__trid:
            return
        if self.__trid != trid:
            gLogger.error("OOps. trid's don't match. This shouldn't happen!", "(%s vs %s)" % (self.__trid, trid))
            return S_ERROR("OOOPS")
        self.__trid = False
        try:
            self.__transport.close()
        except Exception:
            pass
        for cb in self.__specialCallbacks["drop"]:
            try:
                cb(self)
            except Exception:
                gLogger.exception("Exception while processing disconnect callbacks")

    def __cbRecvMsg(self, trid, msgObj):
        msgName = msgObj.getName()
        msgObj.setMsgClient(self)
        for cb in self.__specialCallbacks["msg"]:
            try:
                result = cb(self, msgObj)
                if not isReturnStructure(result):
                    gLogger.error("Callback for message does not return S_OK/S_ERROR", msgObj.getName())
                    return S_ERROR("No response")
                if not result["OK"]:
                    return result
                # If no specific callback but a generic one, return the generic one
                if msgName not in self.__callbacks:
                    return result
            except Exception:
                gLogger.exception("Exception while processing callbacks", msgObj.getName())
        if msgName not in self.__callbacks:
            return S_ERROR("Unexpected message")
        try:
            result = self.__callbacks[msgName](msgObj)
            if not isReturnStructure(result):
                gLogger.error("Callback for message does not return S_OK/S_ERROR", msgName)
                return S_ERROR("No response")
            return result
        except Exception:
            gLogger.exception("Exception while processing callbacks", msgName)
        return S_ERROR("No response")

    def getTrid(self):
        return self.__trid

    def sendMessage(self, msgObj):
        if not self.__trid:
            result = self.connect()
            if not result["OK"]:
                gLogger.error("Failed connect for sending", "%r" % msgObj)
                return result
        return self.__msgBroker.sendMessage(self.__trid, msgObj)

    def subscribeToAllMessages(self, cbFunction):
        if not callable(cbFunction):
            return S_ERROR("%s is not callable" % cbFunction)
        self.__specialCallbacks["msg"].append(cbFunction)
        return S_OK()

    def subscribeToMessage(self, msgName, cbFunction):
        if not callable(cbFunction):
            return S_ERROR("%s is not callable" % cbFunction)
        self.__callbacks[msgName] = cbFunction
        return S_OK()

    def subscribeToDisconnect(self, cbFunction):
        if not callable(cbFunction):
            return S_ERROR("%s is not callable" % cbFunction)
        self.__specialCallbacks["drop"].append(cbFunction)
        return S_OK()

    def clearSubscription(self, msgName):
        try:
            del self.__callbacks[msgName]
        except KeyError:
            return False
        return True

    def disconnect(self):
        trid = self.__trid
        self.__trid = False
        self.__msgBroker.removeTransport(trid)



"""Test for WMS clients."""
# pylint: disable=protected-access, missing-docstring, invalid-name

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import pytest

from mock import MagicMock

from DIRAC import S_OK, S_ERROR
from DIRAC.WorkloadManagementSystem.Client.DownloadInputData import DownloadInputData

MODULE_NAME = "DIRAC.WorkloadManagementSystem.Client.DownloadInputData"

# pylint: disable=redefined-outer-name, unused-argument


@pytest.fixture(autouse=True)
def setUpLogger():
    from DIRAC import gLogger

    gLogger.setLevel("DEBUG")


@pytest.fixture
def mockSE(mocker):
    mockObjectSE = MagicMock()
    mockObjectSE.getFileMetadata.return_value = S_OK(
        {"Successful": {"/a/lfn/1.txt": {"Cached": 1, "Accessible": 1}}, "Failed": {}}
    )
    mockObjectSE.getFile.return_value = S_OK({"Successful": {"/a/lfn/1.txt": {}}, "Failed": {}})
    mockObjectSE.getStatus.return_value = S_OK({"Read": True, "DiskSE": True})
    mockObjectSE.status.return_value = {"Read": True, "DiskSE": True}

    theMockSE = MagicMock()
    theMockSE.return_value = mockObjectSE
    mocker.patch(MODULE_NAME + ".StorageElement", new=theMockSE)
    return theMockSE


@pytest.fixture
def dli():
    theDLI = DownloadInputData(
        {
            "InputData": [],
            "Configuration": {"LocalSEList": ["SE_Local"]},
            "InputDataDirectory": "CWD",
            "FileCatalog": S_OK(
                {
                    "Successful": {
                        "/a/lfn/1.txt": {
                            "Size": 10,
                            "GUID": "aGUID",
                            "SE_Local": "",
                            "SE_Remote": "",
                            "SE_Bad": "",
                            "SE_Tape": "",
                        }
                    }
                }
            ),
        }
    )
    theDLI.availableSEs = ["SE_Local", "SE_Remote"]
    return theDLI


@pytest.fixture
def osPathExists(mocker):
    osPathMock = MagicMock(return_value=True)
    mocker.patch("%s.os.path.exists" % MODULE_NAME, new=osPathMock)
    return osPathMock


def test_DLIDownloadFromSE_fail(dli, mockSE, osPathExists):
    osPathExists.return_value = False
    res = dli._downloadFromSE("/a/lfn/1.txt", "mySE", {"mySE": []}, "aGuid")
    assert not res["OK"]


def test_DLIDownloadFromSE_local(dli, mockSE, osPathExists):
    res = dli._downloadFromSE("/a/lfn/1.txt", "mySE", {"mySE": []}, "aGuid")
    assert res["OK"], res
    assert res["Value"]["protocol"] == "LocalData"


def test_DLIDownloadFromSE_realDown(dli, mockSE, osPathExists):
    osPathExists.side_effect = (False, False, True)
    res = dli._downloadFromSE("/a/lfn/1.txt", "mySE", {"mySE": []}, "aGuid")
    assert res["OK"], res.get("Message", "No Error")
    assert res["Value"]["protocol"] == "Downloaded"
    assert "path" in res["Value"]
    assert res["Value"]["path"].endswith("1.txt")


def test_DLIDownloadFromSE_realDown_Fail(dli, mockSE, osPathExists):
    mockObjectSE = mockSE.return_value
    mockObjectSE.getFile.return_value = S_OK({"Successful": {}, "Failed": {"/a/lfn/1.txt": {}}})
    mockObjectSE.getStatus.return_value = S_OK({"Read": True, "DiskSE": True})

    osPathExists.return_value = False
    res = dli._downloadFromSE("/a/lfn/1.txt", "mySE", {"mySE": []}, "aGuid")
    assert not res["OK"], res.get("Message", "No Error")


def test_DLIDownloadFromBestSE_isLocal(dli, mockSE, osPathExists):
    res = dli._downloadFromBestSE("/a/lfn/1.txt", {"mySE": []}, "aGuid")
    assert res["OK"], res.get("Message", "No error")
    assert res["Value"]["protocol"] == "LocalData"


def test_DLIDownloadFromBestSE_Fail(dli, mockSE, osPathExists):
    osPathExists.return_value = False
    res = dli._downloadFromBestSE("/a/lfn/1.txt", {"mySE": []}, "aGuid")
    assert not res["OK"]


def test_DLI_execute(dli, mockSE):
    dli._downloadFromSE = MagicMock(return_value=S_OK({"path": "/local/path/1.txt"}))
    res = dli.execute(dataToResolve=["/a/lfn/1.txt"])
    assert res["OK"]
    assert not res["Value"]["Failed"]
    assert "/a/lfn/1.txt" in res["Value"]["Successful"], res


def test_DLI_execute_getFileMetadata_Fails(dli, mockSE):
    """When getFileMetadata fails for the first SE, we should fall back to the second."""
    mockObjectSE = mockSE.return_value
    mockObjectSE.getFileMetadata.return_value = S_OK(
        {"Successful": {}, "Failed": {"/a/lfn/1.txt": "Error Getting MetaData"}}
    )

    dli._downloadFromSE = MagicMock(return_value=S_OK({"path": "/local/path/1.txt"}))
    dli._isCache = MagicMock(return_value=True)
    res = dli.execute(dataToResolve=["/a/lfn/1.txt"])
    assert res["OK"]
    assert not res["Value"]["Failed"]
    assert "/a/lfn/1.txt" in res["Value"]["Successful"], res
    assert res["Value"]["Successful"]["/a/lfn/1.txt"]["path"] == "/local/path/1.txt", res


def test_DLI_execute_getFileMetadata_Lost(dli, mockSE):
    """When getFileMetadata fails for the first SE, we should fall back to the second."""
    mockObjectSE = mockSE.return_value
    mockObjectSE.getFileMetadata.return_value = S_OK(
        {
            "Successful": {
                "/a/lfn/1.txt": {
                    "Cached": 1,
                    "Accessible": 0,
                    "Lost": True,
                }
            },
            "Failed": {},
        }
    )
    dli._downloadFromSE = MagicMock(return_value=S_OK({"path": "/local/path/1.txt"}))
    dli._isCache = MagicMock(return_value=True)
    res = dli.execute(dataToResolve=["/a/lfn/1.txt"])
    assert res["OK"]
    assert not res["Value"]["Failed"]
    assert "/a/lfn/1.txt" in res["Value"]["Successful"], res
    assert res["Value"]["Successful"]["/a/lfn/1.txt"]["path"] == "/local/path/1.txt", res


def test_DLI_execute_getFileMetadata_Unavailable(dli, mockSE):
    """When getFileMetadata fails for the first SE, we should fall back to the second."""
    mockObjectSE = mockSE.return_value
    mockObjectSE.getFileMetadata.return_value = S_OK(
        {
            "Successful": {
                "/a/lfn/1.txt": {
                    "Cached": 0,
                    "Accessible": 0,
                    "Unavailable": True,
                }
            },
            "Failed": {},
        }
    )
    dli._downloadFromSE = MagicMock(return_value=S_OK({"path": "/local/path/1.txt"}))
    dli._isCache = MagicMock(return_value=True)
    res = dli.execute(dataToResolve=["/a/lfn/1.txt"])
    assert res["OK"]
    assert not res["Value"]["Failed"]
    assert "/a/lfn/1.txt" in res["Value"]["Successful"], res
    assert res["Value"]["Successful"]["/a/lfn/1.txt"]["path"] == "/local/path/1.txt", res


def test_DLI_execute_getFileMetadata_Cached(dli, mockSE):
    """When getFileMetadata fails for the first SE, we should fall back to the second."""
    mockObjectSE = mockSE.return_value
    mockObjectSE.getFileMetadata.return_value = S_OK(
        {
            "Successful": {
                "/a/lfn/1.txt": {
                    "Cached": 0,
                    "Accessible": 0,
                    "Unavailable": False,
                }
            },
            "Failed": {},
        }
    )
    dli._downloadFromSE = MagicMock(return_value=S_OK({"path": "/local/path/1.txt"}))
    dli._isCache = MagicMock(return_value=True)
    res = dli.execute(dataToResolve=["/a/lfn/1.txt"])
    assert res["OK"]
    assert not res["Value"]["Failed"]
    assert "/a/lfn/1.txt" in res["Value"]["Successful"], res
    assert res["Value"]["Successful"]["/a/lfn/1.txt"]["path"] == "/local/path/1.txt", res


def test_DLI_execute_FirstDownFailed(dli, mockSE):
    """When getFileMetadata fails for the first SE, we should fall back to the second."""
    mockObjectSE = mockSE.return_value
    mockObjectSE.getFileMetadata.return_value = S_OK(
        {"Successful": {"/a/lfn/1.txt": {"Cached": 1, "Accessible": 0}}, "Failed": {}}
    )
    dli._downloadFromSE = MagicMock(side_effect=[S_ERROR("Failed to down"), S_OK({"path": "/local/path/1.txt"})])
    dli._isCache = MagicMock(return_value=True)
    res = dli.execute(dataToResolve=["/a/lfn/1.txt"])
    assert res["OK"]
    assert not res["Value"]["Failed"]
    assert "/a/lfn/1.txt" in res["Value"]["Successful"], res
    assert res["Value"]["Successful"]["/a/lfn/1.txt"]["path"] == "/local/path/1.txt", res


def test_DLI_execute_AllDownFailed(dli, mockSE):
    """When getFileMetadata fails for the first SE, we should fall back to the second."""
    mockObjectSE = mockSE.return_value
    mockObjectSE.getFileMetadata.return_value = S_OK(
        {"Successful": {"/a/lfn/1.txt": {"Cached": 1, "Accessible": 0}}, "Failed": {}}
    )
    dli._downloadFromSE = MagicMock(return_value=S_ERROR("Failed to down"))
    dli._isCache = MagicMock(return_value=True)
    res = dli.execute(dataToResolve=["/a/lfn/1.txt"])
    assert res["OK"]
    assert res["Value"]["Failed"]
    assert "/a/lfn/1.txt" in res["Value"]["Failed"], res
    assert res["Value"]["Failed"][0] == "/a/lfn/1.txt", res


def test_DLI_execute_NoLocal(dli, mockSE):
    """Data only at the remote SE."""
    dli = DownloadInputData(
        {
            "InputData": [],
            "Configuration": {"LocalSEList": ["SE_Local", "SE_Tape"]},
            "InputDataDirectory": "CWD",
            "FileCatalog": S_OK(
                {
                    "Successful": {
                        "/a/lfn/1.txt": {
                            "Size": 10,
                            "GUID": "aGUID",
                            "SE_Tape": "",
                        }
                    }
                }
            ),
        }
    )
    dli.availableSEs = ["SE_Local", "SE_Remote", "SE_Tape"]
    mockObjectSE = mockSE.return_value
    statDict = {"Read": True, "TapeSE": True, "DiskSE": False}
    mockObjectSE.getStatus.return_value = S_OK(statDict)
    mockObjectSE.status.return_value = statDict
    dli._downloadFromSE = MagicMock(return_value=S_ERROR("Failed to down"))
    dli._isCache = MagicMock(return_value=True)

    res = dli.execute(dataToResolve=["/a/lfn/1.txt"])
    assert res["OK"]
    assert res["Value"]["Failed"]
    assert "/a/lfn/1.txt" in res["Value"]["Failed"], res
    assert res["Value"]["Failed"][0] == "/a/lfn/1.txt", res



"""
Update the transformation files of active transformations,
given an InputDataQuery fetched from the Transformation Service.

Possibility to speedup the query time by only fetching files that were added since the last iteration.
Use the CS option RefreshOnly (False by default) and set the DateKey (empty by default) to the meta data
key set in the DIRAC FileCatalog.

The following options can be set for the InputDataAgent.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN InputDataAgent
  :end-before: ##END
  :dedent: 2
  :caption: InputDataAgent options
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
import datetime

from errno import ENOENT

from DIRAC import S_OK
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.DErrno import cmpError
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations

__RCSID__ = "$Id$"

AGENT_NAME = "Transformation/InputDataAgent"


class InputDataAgent(AgentModule):
    def __init__(self, *args, **kwargs):
        """c'tor"""
        AgentModule.__init__(self, *args, **kwargs)

        self.fileLog = {}
        self.timeLog = {}
        self.fullTimeLog = {}

        self.pollingTime = self.am_getOption("PollingTime", 120)
        self.fullUpdatePeriod = self.am_getOption("FullUpdatePeriod", 86400)
        self.refreshonly = self.am_getOption("RefreshOnly", False)
        self.dateKey = self.am_getOption("DateKey", None)

        self.transClient = TransformationClient()
        self.metadataClient = FileCatalogClient()
        self.transformationTypes = None

    #############################################################################
    def initialize(self):
        """Make the necessary initializations"""
        gMonitor.registerActivity("Iteration", "Agent Loops", AGENT_NAME, "Loops/min", gMonitor.OP_SUM)
        agentTSTypes = self.am_getOption("TransformationTypes", [])
        if agentTSTypes:
            self.transformationTypes = sorted(agentTSTypes)
        else:
            dataProc = Operations().getValue("Transformations/DataProcessing", ["MCSimulation", "Merge"])
            dataManip = Operations().getValue("Transformations/DataManipulation", ["Replication", "Removal"])
            self.transformationTypes = sorted(dataProc + dataManip)
        extendables = Operations().getValue("Transformations/ExtendableTransfTypes", [])
        if extendables:
            for extendable in extendables:
                if extendable in self.transformationTypes:
                    self.transformationTypes.remove(extendable)
                    # This is because the Extendables do not use this Agent (have no Input data query)

        return S_OK()

    ##############################################################################
    def execute(self):
        """Main execution method"""

        gMonitor.addMark("Iteration", 1)
        # Get all the transformations
        result = self.transClient.getTransformations({"Status": "Active", "Type": self.transformationTypes})
        if not result["OK"]:
            self.log.error("InputDataAgent.execute: Failed to get transformations.", result["Message"])
            return S_OK()

        # Process each transformation
        for transDict in result["Value"]:
            transID = int(transDict["TransformationID"])
            # res = self.transClient.getTransformationInputDataQuery( transID )
            res = self.transClient.getTransformationMetaQuery(transID, "Input")
            if not res["OK"]:
                if cmpError(res, ENOENT):
                    self.log.info("InputDataAgent.execute: No input data query found for transformation", transID)
                else:
                    self.log.error(
                        "InputDataAgent.execute: Failed to get input data query",
                        "for %d: %s" % (transID, res["Message"]),
                    )
                continue
            inputDataQuery = res["Value"]

            if self.refreshonly:
                # Determine the correct time stamp to use for this transformation
                if transID in self.timeLog:
                    if transID in self.fullTimeLog:
                        # If it is more than a day since the last reduced query, make a full query just in case
                        if (datetime.datetime.utcnow() - self.fullTimeLog[transID]) < datetime.timedelta(
                            seconds=self.fullUpdatePeriod
                        ):
                            timeStamp = self.timeLog[transID]
                            if self.dateKey:
                                inputDataQuery[self.dateKey] = (timeStamp - datetime.timedelta(seconds=10)).strftime(
                                    "%Y-%m-%d %H:%M:%S"
                                )
                            else:
                                self.log.error("DateKey was not set in the CS, cannot use the RefreshOnly")
                        else:
                            self.fullTimeLog[transID] = datetime.datetime.utcnow()
                self.timeLog[transID] = datetime.datetime.utcnow()
                if transID not in self.fullTimeLog:
                    self.fullTimeLog[transID] = datetime.datetime.utcnow()

            # Perform the query to the metadata catalog
            self.log.verbose("Using input data query for transformation", "%d: %s" % (transID, str(inputDataQuery)))
            start = time.time()
            result = self.metadataClient.findFilesByMetadata(inputDataQuery)
            rtime = time.time() - start
            self.log.verbose("Metadata catalog query time", ": %.2f seconds." % (rtime))
            if not result["OK"]:
                self.log.error(
                    "InputDataAgent.execute: Failed to get response from the metadata catalog", result["Message"]
                )
                continue
            lfnList = result["Value"]

            # Check if the number of files has changed since the last cycle
            nlfns = len(lfnList)
            self.log.info(
                "files returned for transformation from the metadata catalog: ", "%d -> %d" % (int(transID), nlfns)
            )
            if nlfns == self.fileLog.get(transID):
                self.log.verbose("No new files in metadata catalog since last check")
            self.fileLog[transID] = nlfns

            # Add any new files to the transformation
            addedLfns = []
            if lfnList:
                self.log.verbose("Processing lfns for transformation:", "%d -> %d" % (transID, len(lfnList)))
                # Add the files to the transformation
                self.log.verbose("Adding lfns for transformation:", "%d -> %d" % (transID, len(lfnList)))
                result = self.transClient.addFilesToTransformation(transID, sorted(lfnList))
                if not result["OK"]:
                    self.log.warn("InputDataAgent.execute: failed to add lfns to transformation", result["Message"])
                    self.fileLog[transID] = 0
                else:
                    if result["Value"]["Failed"]:
                        for lfn, error in res["Value"]["Failed"].items():
                            self.log.warn(
                                "InputDataAgent.execute: Failed to add to transformation:", "%s: %s" % (lfn, error)
                            )
                    if result["Value"]["Successful"]:
                        for lfn, status in result["Value"]["Successful"].items():
                            if status == "Added":
                                addedLfns.append(lfn)
                        self.log.info("InputDataAgent.execute: Added files to transformation", "(%d)" % len(addedLfns))

        return S_OK()



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import unittest

from DIRAC.Core.Workflow.Module import ModuleDefinition
from DIRAC.Core.Workflow.Step import StepDefinition
from DIRAC.Interfaces.API.Job import Job
from DIRAC.Workflow.Utilities.Utils import getStepDefinition, getStepCPUTimes

#############################################################################


class UtilitiesTestCase(unittest.TestCase):
    """Base class"""

    def setUp(self):

        self.job = Job()
        pass


class UtilsSuccess(UtilitiesTestCase):
    def test__getStepDefinition(self):
        importLine = """
from DIRAC.Workflow.Modules.<MODULE> import <MODULE>
"""
        # modules
        gaudiApp = ModuleDefinition("Script")
        body = importLine.replace("<MODULE>", "Script")
        gaudiApp.setDescription(
            getattr(
                __import__("%s.%s" % ("DIRAC.Workflow.Modules", "Script"), globals(), locals(), ["__doc__"]), "__doc__"
            )
        )
        gaudiApp.setBody(body)

        genBKReport = ModuleDefinition("FailoverRequest")
        body = importLine.replace("<MODULE>", "FailoverRequest")
        genBKReport.setDescription(
            getattr(
                __import__("%s.%s" % ("DIRAC.Workflow.Modules", "FailoverRequest"), globals(), locals(), ["__doc__"]),
                "__doc__",
            )
        )
        genBKReport.setBody(body)

        # step
        appDefn = StepDefinition("App_Step")
        appDefn.addModule(gaudiApp)
        appDefn.createModuleInstance("Script", "Script")
        appDefn.addModule(genBKReport)
        appDefn.createModuleInstance("FailoverRequest", "FailoverRequest")

        appDefn.addParameterLinked(gaudiApp.parameters)

        stepDef = getStepDefinition("App_Step", ["Script", "FailoverRequest"])

        self.assertTrue(str(appDefn) == str(stepDef))

        self.job._addParameter(appDefn, "name", "type", "value", "desc")
        self.job._addParameter(appDefn, "name1", "type1", "value1", "desc1")

        stepDef = getStepDefinition(
            "App_Step",
            ["Script", "FailoverRequest"],
            parametersList=[["name", "type", "value", "desc"], ["name1", "type1", "value1", "desc1"]],
        )

        self.assertTrue(str(appDefn) == str(stepDef))

    def test_getStepCPUTimes(self):
        execT, cpuT = getStepCPUTimes({})
        self.assertEqual(execT, 0)
        self.assertEqual(cpuT, 0)
        execT, cpuT = getStepCPUTimes({"StartTime": 0, "StartStats": (0, 0, 0, 0, 0)})
        print(execT, cpuT)


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(UtilitiesTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(UtilsSuccess))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



""" MQProducer
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import gLogger
from DIRAC.Resources.MessageQueue.Utilities import getDestinationAddress, getMQService


class MQProducer(object):
    def __init__(self, mqManager, mqURI, producerId):
        self._connectionManager = mqManager
        self._mqURI = mqURI
        self._destination = getDestinationAddress(self._mqURI)
        self._mqService = getMQService(self._mqURI)
        self._id = producerId
        self.log = gLogger.getSubLogger(self.__class__.__name__)

    def put(self, msg):
        result = self._connectionManager.getConnector(self._mqService)
        if result["OK"]:
            connector = result["Value"]
            return connector.put(message=msg, parameters={"destination": self._destination})
        return result

    def close(self):
        """Function closes the connection for this client.
            The producer id is removed from the connection storage.
            It is not guaranteed that the connection will be
            removed cause other messengers can be still using it.

        Returns:
          S_OK or S_ERROR: Error appears in case if the connection was already
            closed for this producer.
        """
        return self._connectionManager.stopConnection(mqURI=self._mqURI, messengerId=self._id)



""" File catalog class. This is a simple dispatcher for the file catalog plug-ins.
    It ensures that all operations are performed on the desired catalogs.

    The File Catalog plug-ins are supposed to implement a certain number of methods of
    the File Catalog interface. The names of the implemented methods classified in
    "read", "write" and "no_lfn" categories are reported by each plug-in using
    getInterfaceMethods() call. The File Catalog collects and memorizes all the
    method names from all the plug-ins and in each category. Only calls to these
    methods will be attempted.

    One plug-in can be declared to be a Master in the CS. If the Master plug-in is
    declared it must implement all the methods collected in the "write" category.
    When the FileCatalog is called with a given "write" method name, the Master plugin
    is called first. If it fails, no other plug-in is called to preserve consistency
    in the states of different catalogs. If no Master plug-in is declared, all the
    plug-ins are called (in case they implement the method) for the "write" methods.

    For the "read" methods plug-ins are called one by one, starting with the Master
    plug-in if declared, until getting a successful result.

    Most of the catalog plug-in methods are taking the first argument which represents
    the required LFNS. The LFNs argument can have one of the following forms:

    - string: just a single LFN itself
    - list: a list of LFN strings
    - dictionary: the keys are LFN strings, the values are LFN specific parameters
                  needed by the method

    All the methods taking the first LFNs argument are returning a standard DIRAC
    bulk result structure. If the call is successful, the Successful and Failed
    dictionaries have LFNs as keys and specific results of operation as values.
    The LFNs argument before calling these methods is checked to conform to the
    convention above and modified to take the "dictionary" form. The original LFN names
    are memorized and restored in the final result.

    Some methods implemented by plug-ins do not have LFNs as the first argument.
    The names of those methods are reported by the plug-ins as "no_lfn" methods
    in the getInterfaceMethods() call. For those methods there is obviously no
    additional check of the structure of the LFNs argument and no corresponding
    processing of the results.

    For the actual methods that can be called vie the File Catalog object, see
    the documentation of the respective FileCatalog plug-ins ( client classes )

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import re

from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.Core.Utilities import DErrno
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Security.ProxyInfo import getVOfromProxyGroup
from DIRAC.Resources.Catalog.Utilities import checkArgumentFormat
from DIRAC.Resources.Catalog.FileCatalogFactory import FileCatalogFactory
from DIRAC.Resources.Catalog.FCConditionParser import FCConditionParser


class FileCatalog(object):
    def __init__(self, catalogs=None, vo=None):
        """Default constructor"""
        self.valid = True
        self.timeout = 180

        self.ro_methods = set()
        self.write_methods = set()
        self.no_lfn_methods = set()

        self.readCatalogs = []
        self.writeCatalogs = []
        self.rootConfigPath = "/Resources/FileCatalogs"
        self.vo = vo if vo else getVOfromProxyGroup().get("Value", None)
        self.log = gLogger.getSubLogger("FileCatalog")

        self.opHelper = Operations(vo=self.vo)

        catalogList = []
        if isinstance(catalogs, six.string_types):
            catalogList = [catalogs]
        elif isinstance(catalogs, (list, tuple)):
            catalogList = list(catalogs)

        if catalogList:
            result = self._getEligibleCatalogs()
            if not result["OK"]:
                self.log.error("Failed to get eligible catalog")
                return
            eligibleFileCatalogs = result["Value"]
            catalogCheck = True
            for catalog in catalogList:
                if catalog not in eligibleFileCatalogs:
                    self.log.error("Specified catalog is not eligible", catalog)
                    catalogCheck = False
            if catalogCheck:
                result = self._getSelectedCatalogs(catalogList)
            else:
                result = S_ERROR("Specified catalog is not eligible")
        else:
            result = self._getCatalogs()
        if not result["OK"]:
            self.log.error("Failed to create catalog objects")
            self.valid = False
        elif (len(self.readCatalogs) == 0) and (len(self.writeCatalogs) == 0):
            self.log.error("No catalog object created")
            self.valid = False

        result = self.getMasterCatalogNames()
        masterCatalogs = result["Value"]
        # There can not be more than one master catalog
        haveMaster = False
        if len(masterCatalogs) > 1:
            self.log.error("More than one master catalog created")
            self.valid = False
        elif len(masterCatalogs) == 1:
            haveMaster = True

        # Get the list of write methods
        if haveMaster:
            # All the write methods must be present in the master
            _catalogName, oCatalog, _master = self.writeCatalogs[0]
            _roList, writeList, nolfnList = oCatalog.getInterfaceMethods()
            self.write_methods.update(writeList)
            self.no_lfn_methods.update(nolfnList)
        else:
            for _catalogName, oCatalog, _master in self.writeCatalogs:
                _roList, writeList, nolfnList = oCatalog.getInterfaceMethods()
                self.write_methods.update(writeList)
                self.no_lfn_methods.update(nolfnList)

        # Get the list of read methods
        for _catalogName, oCatalog, _master in self.readCatalogs:
            roList, _writeList, nolfnList = oCatalog.getInterfaceMethods()
            self.ro_methods.update(roList)
            self.no_lfn_methods.update(nolfnList)

        self.condParser = FCConditionParser(vo=self.vo, ro_methods=self.ro_methods)

    def isOK(self):
        return self.valid

    def getReadCatalogs(self):
        return self.readCatalogs

    def getWriteCatalogs(self):
        return self.writeCatalogs

    def getMasterCatalogNames(self):
        """Returns the list of names of the Master catalogs"""

        masterNames = [catalogName for catalogName, oCatalog, master in self.writeCatalogs if master]
        return S_OK(masterNames)

    def __getattr__(self, name):
        self.call = name
        if name in self.write_methods:
            return self.w_execute
        elif name in self.ro_methods:
            return self.r_execute
        else:
            raise AttributeError

    def w_execute(self, *parms, **kws):
        """Write method executor.

        If one of the LFNs given as input does not pass a condition defined for the
        master catalog, we return S_ERROR without trying anything else

        :param fcConditions: either a dict or a string, to be propagated to the FCConditionParser

                             * If it is a string, it is given for all catalogs
                             * If it is a dict, it has to be { catalogName: condition}, and only
                                  the specific condition for the catalog will be given

        .. warning ::

          If the method is a write no_lfn method, then the return value are completely different.
          We only return the result of the master catalog


        """
        successful = {}
        failed = {}
        failedCatalogs = {}
        successfulCatalogs = {}

        specialConditions = kws.pop("fcConditions") if "fcConditions" in kws else None

        allLfns = []
        lfnMapDict = {}
        masterResult = {}
        parms1 = []
        if self.call not in self.no_lfn_methods:
            fileInfo = parms[0]
            result = checkArgumentFormat(fileInfo, generateMap=True)
            if not result["OK"]:
                return result
            fileInfo, lfnMapDict = result["Value"]
            # No need to check the LFNs again in the clients
            kws["LFNChecking"] = False
            allLfns = list(fileInfo)
            parms1 = parms[1:]

        for catalogName, oCatalog, master in self.writeCatalogs:

            # Skip if the method is not implemented in this catalog
            # NOTE: it is impossible for the master since the write method list is populated
            # only from the master catalog, and if the method is not there, __getattr__
            # would raise an exception
            if not oCatalog.hasCatalogMethod(self.call):
                continue

            method = getattr(oCatalog, self.call)

            if self.call in self.no_lfn_methods:
                result = method(*parms, **kws)
            else:
                if isinstance(specialConditions, dict):
                    condition = specialConditions.get(catalogName)
                else:
                    condition = specialConditions
                # Check whether this catalog should be used for this method
                res = self.condParser(catalogName, self.call, fileInfo, condition=condition)
                # condParser never returns S_ERROR
                condEvals = res["Value"]["Successful"]
                # For a master catalog, ALL the lfns should be valid
                if master:
                    if any([not valid for valid in condEvals.values()]):
                        gLogger.error("The master catalog is not valid for some LFNS", condEvals)
                        return S_ERROR("The master catalog is not valid for some LFNS %s" % condEvals)

                validLFNs = dict((lfn, fileInfo[lfn]) for lfn in condEvals if condEvals[lfn])

                # We can skip the execution without worry,
                # since at this level it is for sure not a master catalog
                if not validLFNs:
                    gLogger.debug("No valid LFN, skipping the call")
                    continue

                invalidLFNs = [lfn for lfn in condEvals if not condEvals[lfn]]

                if invalidLFNs:
                    gLogger.debug(
                        "Some LFNs are not valid for operation '%s' on catalog '%s' : %s"
                        % (self.call, catalogName, invalidLFNs)
                    )

                result = method(validLFNs, *parms1, **kws)

            if master:
                masterResult = result

            if not result["OK"]:
                if master:
                    # If this is the master catalog and it fails we don't want to continue with the other catalogs
                    self.log.error(
                        "Failed to execute call on master catalog",
                        "%s on %s: %s" % (self.call, catalogName, result["Message"]),
                    )
                    return result
                else:
                    # Otherwise we keep the failed catalogs so we can update their state later
                    failedCatalogs[catalogName] = result["Message"]
            else:
                successfulCatalogs[catalogName] = result["Value"]

            if allLfns:
                if result["OK"]:
                    for lfn, message in result["Value"]["Failed"].items():
                        # Save the error message for the failed operations
                        failed.setdefault(lfn, {})[catalogName] = message
                        if master:
                            # If this is the master catalog then we should not attempt the operation on other catalogs
                            fileInfo.pop(lfn, None)
                    for lfn, result in result["Value"]["Successful"].items():
                        # Save the result return for each file for the successful operations
                        successful.setdefault(lfn, {})[catalogName] = result

        if allLfns:
            # This recovers the states of the files that completely failed i.e. when S_ERROR is returned by a catalog
            for catalogName, errorMessage in failedCatalogs.items():
                for lfn in allLfns:
                    failed.setdefault(lfn, {})[catalogName] = errorMessage
            # Restore original lfns if they were changed by normalization
            if lfnMapDict:
                for lfn in list(failed):
                    failed[lfnMapDict.get(lfn, lfn)] = failed.pop(lfn)
                for lfn in list(successful):
                    successful[lfnMapDict.get(lfn, lfn)] = successful.pop(lfn)
            resDict = {"Failed": failed, "Successful": successful}
            return S_OK(resDict)
        else:
            # FIXME: Return just master result here. This is temporary as more detailed
            # per catalog result needs multiple fixes in various client calls
            return masterResult

    def r_execute(self, *parms, **kws):
        """Read method executor."""
        successful = {}
        failed = {}
        for _catalogName, oCatalog, _master in self.readCatalogs:

            # Skip if the method is not implemented in this catalog
            if not oCatalog.hasCatalogMethod(self.call):
                continue

            method = getattr(oCatalog, self.call)
            res = method(*parms, **kws)
            if res["OK"]:
                if "Successful" in res["Value"]:
                    for key, item in res["Value"]["Successful"].items():
                        successful.setdefault(key, item)
                        failed.pop(key, None)
                    for key, item in res["Value"]["Failed"].items():
                        if key not in successful:
                            failed[key] = item
                else:
                    return res
        if not successful and not failed:
            return S_ERROR(DErrno.EFCERR, "Failed to perform %s from any catalog" % self.call)
        return S_OK({"Failed": failed, "Successful": successful})

    ###########################################################################################
    #
    # Below is the method for obtaining the objects instantiated for a provided catalogue configuration
    #

    def addCatalog(self, catalogName, mode="Write", master=False):
        """Add a new catalog with catalogName to the pool of catalogs in mode:
        "Read","Write" or "ReadWrite"
        """

        result = self._generateCatalogObject(catalogName)
        if not result["OK"]:
            return result

        oCatalog = result["Value"]
        if mode.lower().find("read") != -1:
            self.readCatalogs.append((catalogName, oCatalog, master))
        if mode.lower().find("write") != -1:
            self.writeCatalogs.append((catalogName, oCatalog, master))

        return S_OK()

    def removeCatalog(self, catalogName):
        """Remove the specified catalog from the internal pool"""

        catalog_removed = False

        for i in range(len(self.readCatalogs)):
            catalog, _object, _master = self.readCatalogs[i]
            if catalog == catalogName:
                del self.readCatalogs[i]
                catalog_removed = True
                break
        for i in range(len(self.writeCatalogs)):
            catalog, _object, _master = self.writeCatalogs[i]
            if catalog == catalogName:
                del self.writeCatalogs[i]
                catalog_removed = True
                break

        if catalog_removed:
            return S_OK()
        else:
            return S_OK("Catalog does not exist")

    def _getSelectedCatalogs(self, desiredCatalogs):
        for catalogName in desiredCatalogs:
            result = self._getCatalogConfigDetails(catalogName)
            if not result["OK"]:
                return result
            catalogConfig = result["Value"]
            result = self._generateCatalogObject(catalogName)
            if not result["OK"]:
                return result
            oCatalog = result["Value"]
            if re.search("Read", catalogConfig["AccessType"]):
                if catalogConfig["Master"]:
                    self.readCatalogs.insert(0, (catalogName, oCatalog, catalogConfig["Master"]))
                else:
                    self.readCatalogs.append((catalogName, oCatalog, catalogConfig["Master"]))
            if re.search("Write", catalogConfig["AccessType"]):
                if catalogConfig["Master"]:
                    self.writeCatalogs.insert(0, (catalogName, oCatalog, catalogConfig["Master"]))
                else:
                    self.writeCatalogs.append((catalogName, oCatalog, catalogConfig["Master"]))
        return S_OK()

    def _getEligibleCatalogs(self):
        """Get a list of eligible catalogs

        :return: S_OK/S_ERROR, Value - a list of catalog names
        """
        # First, look in the Operations, if nothing defined look in /Resources for backward compatibility
        fileCatalogs = self.opHelper.getValue("/Services/Catalogs/CatalogList", [])
        if not fileCatalogs:
            result = self.opHelper.getSections("/Services/Catalogs")
            if result["OK"]:
                fileCatalogs = result["Value"]
            else:
                res = gConfig.getSections(self.rootConfigPath, listOrdered=True)
                if not res["OK"]:
                    errStr = "FileCatalog._getEligibleCatalogs: Failed to get file catalog configuration."
                    self.log.error(errStr, res["Message"])
                    return S_ERROR(errStr)
                fileCatalogs = res["Value"]

        return S_OK(fileCatalogs)

    def _getCatalogs(self):
        """Updates self.readCatalogs and self.writeCatalogs with list of catalog objects as found in the CS"""

        # Get the eligible catalogs first
        result = self._getEligibleCatalogs()
        if not result["OK"]:
            return result
        fileCatalogs = result["Value"]

        # Get the catalog objects now
        for catalogName in fileCatalogs:
            res = self._getCatalogConfigDetails(catalogName)
            if not res["OK"]:
                return res
            catalogConfig = res["Value"]
            if catalogConfig["Status"] == "Active":
                res = self._generateCatalogObject(catalogName)
                if not res["OK"]:
                    return res
                oCatalog = res["Value"]
                master = catalogConfig["Master"]
                # If the catalog is read type
                if re.search("Read", catalogConfig["AccessType"]):
                    if master:
                        self.readCatalogs.insert(0, (catalogName, oCatalog, master))
                    else:
                        self.readCatalogs.append((catalogName, oCatalog, master))
                # If the catalog is write type
                if re.search("Write", catalogConfig["AccessType"]):
                    if master:
                        self.writeCatalogs.insert(0, (catalogName, oCatalog, master))
                    else:
                        self.writeCatalogs.append((catalogName, oCatalog, master))
        return S_OK()

    def _getCatalogConfigDetails(self, catalogName):
        # First obtain the options that are available
        catalogConfigPath = "%s/%s" % (self.rootConfigPath, catalogName)
        result = gConfig.getOptionsDict(catalogConfigPath)
        if not result["OK"]:
            errStr = "FileCatalog._getCatalogConfigDetails: Failed to get catalog options."
            self.log.error(errStr, catalogName)
            return S_ERROR(errStr)
        catalogConfig = result["Value"]
        result = self.opHelper.getOptionsDict("/Services/Catalogs/%s" % catalogName)
        if result["OK"]:
            catalogConfig.update(result["Value"])

        # The 'Status' option should be defined (default = 'Active')
        if "Status" not in catalogConfig:
            warnStr = "FileCatalog._getCatalogConfigDetails: 'Status' option not defined."
            self.log.warn(warnStr, catalogName)
            catalogConfig["Status"] = "Active"
        # The 'AccessType' option must be defined
        if "AccessType" not in catalogConfig:
            errStr = "FileCatalog._getCatalogConfigDetails: Required option 'AccessType' not defined."
            self.log.error(errStr, catalogName)
            return S_ERROR(errStr)
        # Anything other than 'True' in the 'Master' option means it is not
        catalogConfig["Master"] = catalogConfig.setdefault("Master", False) == "True"
        return S_OK(catalogConfig)

    def _generateCatalogObject(self, catalogName):
        """Create a file catalog object from its name and CS description"""
        useProxy = gConfig.getValue("/LocalSite/Catalogs/%s/UseProxy" % catalogName, False)
        if not useProxy:
            useProxy = self.opHelper.getValue("/Services/Catalogs/%s/UseProxy" % catalogName, False)
        return FileCatalogFactory().createCatalog(catalogName, useProxy)



"""
Class for management of Stomp MQ connections, e.g. RabbitMQ
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import random
import os
import socket
import ssl
import time
import stomp

from DIRAC.Resources.MessageQueue.MQConnector import MQConnector
from DIRAC.Core.Security import Locations
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.DErrno import EMQUKN, EMQCONN

LOG = gLogger.getSubLogger(__name__)


class StompMQConnector(MQConnector):
    """
    Class for management of message queue connections
    Allows to both send and receive messages from a queue

    When several IPs are behind an alias, we shuffle the ips, and connect to one.
    The others are used as failover by stomp's internals
    """

    # Setting for the reconnection handling by stomp interface.
    # See e.g. the description of Transport class in
    # https://github.com/jasonrbriggs/stomp.py/blob/master/stomp/transport.py

    RECONNECT_SLEEP_INITIAL = 1  # [s]  Initial delay before reattempting to establish a connection.
    RECONNECT_SLEEP_INCREASE = 0.5  # Factor by which sleep delay is increased 0.5 means increase by 50%.
    RECONNECT_SLEEP_MAX = 120  # [s] The maximum delay that can be reached independent of increasing procedure.
    RECONNECT_SLEEP_JITTER = 0.1  # Random factor to add. 0.1 means a random number from 0 to 10% of the current time.
    RECONNECT_ATTEMPTS_MAX = 1e4  # Maximum attempts to reconnect.

    PORT = 61613

    def __init__(self, parameters=None):
        """Standard constructor"""
        super(StompMQConnector, self).__init__(parameters=parameters)
        self.connection = None

        if "DIRAC_DEBUG_STOMP" in os.environ:
            gLogger.enableLogsFromExternalLibs()

    def setupConnection(self, parameters=None):
        """
         Establishes a new connection to a Stomp server, e.g. RabbitMQ

        Args:
          parameters(dict): dictionary with additional MQ parameters if any.
        Returns:
          S_OK/S_ERROR
        """
        log = LOG.getSubLogger("setupConnection")

        if parameters is not None:
            self.parameters.update(parameters)

        # Check that the minimum set of parameters is present
        if not all(p in parameters for p in ("Host", "VHost")):
            return S_ERROR("Input parameters are missing!")

        reconnectSleepInitial = self.parameters.get("ReconnectSleepInitial", StompMQConnector.RECONNECT_SLEEP_INITIAL)
        reconnectSleepIncrease = self.parameters.get(
            "ReconnectSleepIncrease", StompMQConnector.RECONNECT_SLEEP_INCREASE
        )
        reconnectSleepMax = self.parameters.get("ReconnectSleepMax", StompMQConnector.RECONNECT_SLEEP_MAX)
        reconnectSleepJitter = self.parameters.get("ReconnectSleepJitter", StompMQConnector.RECONNECT_SLEEP_JITTER)
        reconnectAttemptsMax = self.parameters.get("ReconnectAttemptsMax", StompMQConnector.RECONNECT_ATTEMPTS_MAX)

        host = self.parameters.get("Host")
        port = self.parameters.get("Port", StompMQConnector.PORT)
        vhost = self.parameters.get("VHost")

        sslVersion = self.parameters.get("SSLVersion")
        hostcert = self.parameters.get("HostCertificate")
        hostkey = self.parameters.get("HostKey")

        connectionArgs = {
            "vhost": vhost,
            "keepalive": True,
            "reconnect_sleep_initial": reconnectSleepInitial,
            "reconnect_sleep_increase": reconnectSleepIncrease,
            "reconnect_sleep_max": reconnectSleepMax,
            "reconnect_sleep_jitter": reconnectSleepJitter,
            "reconnect_attempts_max": reconnectAttemptsMax,
        }

        # We use ssl credentials and not user-password.
        if sslVersion is not None:
            if sslVersion == "TLSv1":
                sslVersion = ssl.PROTOCOL_TLSv1
                # get local key and certificate if not available via configuration
                if not (hostcert or hostkey):
                    paths = Locations.getHostCertificateAndKeyLocation()
                    if not paths:
                        return S_ERROR("Could not find a certificate!")
                    hostcert = paths[0]
                    hostkey = paths[1]
                connectionArgs.update(
                    {"use_ssl": True, "ssl_version": sslVersion, "ssl_key_file": hostkey, "ssl_cert_file": hostcert}
                )
            else:
                return S_ERROR(EMQCONN, "Invalid SSL version provided: %s" % sslVersion)

        try:

            # Get IP addresses of brokers
            # Start with the IPv6, and randomize it
            ipv6_addrInfo = socket.getaddrinfo(host, port, socket.AF_INET6, socket.SOCK_STREAM)
            random.shuffle(ipv6_addrInfo)
            # Same with IPv4
            ipv4_addrInfo = socket.getaddrinfo(host, port, socket.AF_INET, socket.SOCK_STREAM)
            random.shuffle(ipv4_addrInfo)

            # Create the host_port tuples, keeping the ipv6 in front
            host_and_ports = []
            for _family, _socktype, _proto, _canonname, sockaddr in ipv6_addrInfo + ipv4_addrInfo:
                host_and_ports.append((sockaddr[0], sockaddr[1]))

            connectionArgs.update({"host_and_ports": host_and_ports})
            log.debug("Connection args: %s" % str(connectionArgs))
            self.connection = stomp.Connection(**connectionArgs)

        except Exception as e:
            log.debug("Failed setting up connection", repr(e))
            return S_ERROR(EMQCONN, "Failed to setup connection: %s" % e)

        return S_OK("Setup successful")

    def reconnect(self):
        """
        Callback method when a disconnection happens

        :param serverIP: IP of the server disconnected
        """
        log = LOG.getSubLogger("reconnect")
        log.info("Trigger reconnection for broker")
        res = self.connect(self.parameters)
        return res

    def put(self, message, parameters=None):
        """
        Sends a message to the queue
        message contains the body of the message

        Args:
          message(str): string or any json encodable structure.
          parameters(dict): parameters with 'destination' key defined.
        """
        log = LOG.getSubLogger("put")
        destination = parameters.get("destination", "")

        try:
            try:
                self.connection.send(body=json.dumps(message), destination=destination)
            except stomp.exception.StompException:
                self.connect()
                self.connection.send(body=json.dumps(message), destination=destination)
        except Exception as e:
            log.debug("Failed to send message", repr(e))
            return S_ERROR(EMQUKN, "Failed to send message: %s" % repr(e))

        return S_OK("Message sent successfully")

    def connect(self, parameters=None):
        """Call the ~stomp.Connection.connect method for each endpoint

        :param parameters: connection parameter
        """

        log = LOG.getSubLogger("connect")

        # Since I use a dirty trick to know to what IP I am connected,
        # I'd rather not rely too much on it
        remoteIP = "unknown"
        user = self.parameters.get("User")
        password = self.parameters.get("Password")

        for _ in range(10):
            try:
                self.connection.connect(username=user, passcode=password, wait=True)

                if self.connection.is_connected():
                    # Go to the socket of the Stomp to find the remote host
                    try:
                        remoteIP = self.connection.transport.socket.getpeername()[0]
                    except Exception:
                        pass
                    log.info("MQ Connected to %s" % remoteIP)
                    return S_OK("Connected to %s" % remoteIP)
                else:
                    log.warn("Not connected")
            except Exception as e:
                log.error("Failed to connect: %s" % repr(e))

            # Wait a bit before retrying
            time.sleep(5)

        return S_ERROR(EMQCONN, "Failed to connect")

    def disconnect(self, parameters=None):
        """
        Disconnects from the message queue server
        """
        log = LOG.getSubLogger("disconnect")

        try:
            # Indicate to the Listener that we want a disconnection
            listener = self.connection.get_listener("StompListener")
            if listener:
                listener.wantsDisconnect = True

            self.connection.disconnect()
            log.info("Disconnected from broker")
        except Exception as e:
            log.error("Failed to disconnect from broker", repr(e))
            return S_ERROR(EMQUKN, "Failed to disconnect from broker %s" % repr(e))

        return S_OK("Successfully disconnected from broker")

    def subscribe(self, parameters=None):
        log = LOG.getSubLogger("subscribe")

        mId = parameters.get("messengerId", "")
        callback = parameters.get("callback", None)
        dest = parameters.get("destination", "")
        headers = {}
        if self.parameters.get("Persistent", "").lower() in ["true", "yes", "1"]:
            headers = {"persistent": "true"}
        ack = "auto"
        acknowledgement = False
        if self.parameters.get("Acknowledgement", "").lower() in ["true", "yes", "1"]:
            acknowledgement = True
            ack = "client-individual"
        if not callback:
            # Chris 26.02.20
            # If it is an error, why not returning ?!
            log.error("No callback specified!")

        try:
            listener = StompListener(callback, acknowledgement, self.connection, mId, self.connect)
            self.connection.set_listener("StompListener", listener)
            self.connection.subscribe(destination=dest, id=mId, ack=ack, headers=headers)
        except Exception as e:
            log.error("Failed to subscribe: %s" % e)
            return S_ERROR(EMQUKN, "Failed to subscribe to broker: %s" % repr(e))

        return S_OK("Subscription successful")

    def unsubscribe(self, parameters):
        log = LOG.getSubLogger("unsubscribe")

        dest = parameters.get("destination", "")
        mId = parameters.get("messengerId", "")

        try:
            self.connection.unsubscribe(destination=dest, id=mId)
        except Exception as e:
            log.error("Failed to unsubscribe", repr(e))
            return S_ERROR(EMQUKN, "Failed to unsubscribe: %s" % repr(e))

        return S_OK("Successfully unsubscribed from all destinations")


class StompListener(stomp.ConnectionListener):
    """
    Internal listener class responsible for handling new messages and errors.
    """

    def __init__(self, callback, ack, connection, messengerId, connectCallback):
        """
        Initializes the internal listener object

        Args:
          callback: a defaultCallback compatible function.
          ack(bool): if set to true an acknowledgement will be send back to the sender.
          messengerId(str): messenger identifier sent with acknowledgement messages.
          connectCallback: the connect method to call in case of disconnection
        """

        self.log = LOG.getSubLogger("StompListener")
        if not callback:
            self.log.error("Error initializing StompMQConnector!callback is None")
        self.callback = callback
        self.ack = ack
        self.mId = messengerId
        self.connection = connection
        self.connectCallback = connectCallback

        # This boolean is to know whether we effectively
        # want to disconnect or if it is because of a failure
        self.wantsDisconnect = False

    def on_message(self, headers, body):
        """
        Function called upon receiving a message

        :param dict headers: message headers
        :param json body: message body
        """
        result = self.callback(headers, json.loads(body))
        if self.ack:
            if result["OK"]:
                self.connection.ack(headers["message-id"], self.mId)
            else:
                self.connection.nack(headers["message-id"], self.mId)

    def on_error(self, headers, message):
        """Function called when an error happens

        Args:
          headers(dict): message headers.
          body(json): message body.
        """
        self.log.error(message)

    def on_disconnected(self):
        """Callback function called after disconnecting from broker."""
        if not self.wantsDisconnect:
            self.log.warn("Disconnected from broker")
            try:
                res = self.connectCallback()
                if res["OK"]:
                    self.log.info("Reconnection successful to broker")
                else:
                    self.log.error("Error reconnectiong broker", "%s" % res)

            except Exception as e:
                self.log.error("Unexpected error while calling reconnect callback: %s" % e)



""" Collection of user jobs for testing purposes
"""

# pylint: disable=invalid-name

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import time
import errno

from DIRAC import rootPath
from DIRAC.Interfaces.API.Job import Job
from DIRAC.Interfaces.API.Dirac import Dirac

from DIRAC.tests.Utilities.utils import find_all

# parameters

# Common functions


def getJob(jobClass=None):
    if not jobClass:
        jobClass = Job
    oJob = jobClass()
    return oJob


def getDIRAC(diracClass=None):
    if not diracClass:
        diracClass = Dirac
    oDirac = diracClass()
    return oDirac


def baseToAllJobs(jName, jobClass=None):
    print("**********************************************************************************************************")
    print("\n Submitting job ", jName)

    J = getJob(jobClass)
    J.setName(jName)
    J.setCPUTime(17800)
    return J


def endOfAllJobs(J):
    result = getDIRAC().submitJob(J)
    print("Job submission result:", result)
    if result["OK"]:
        print("Submitted with job ID:", result["Value"])

    return result


# List of jobs


def helloWorld():
    """simple hello world job"""

    J = baseToAllJobs("helloWorld")
    try:
        J.setInputSandbox([find_all("exe-script.py", rootPath, "DIRAC/tests/Workflow")[0]])
    except IndexError:  # we are in Jenkins
        J.setInputSandbox([find_all("exe-script.py", os.environ["WORKSPACE"], "DIRAC/tests/Workflow")[0]])
    J.setExecutable("exe-script.py", "", "helloWorld.log")
    return endOfAllJobs(J)


def helloWorldCERN():
    """simple hello world job to CERN"""

    J = baseToAllJobs("helloWorldCERN")
    try:
        J.setInputSandbox([find_all("exe-script.py", rootPath, "DIRAC/tests/Workflow")[0]])
    except IndexError:  # we are in Jenkins
        J.setInputSandbox([find_all("exe-script.py", os.environ["WORKSPACE"], "DIRAC/tests/Workflow")[0]])
    J.setExecutable("exe-script.py", "", "helloWorld.log")
    J.setDestination("LCG.CERN.cern")
    return endOfAllJobs(J)


def helloWorldNCBJ():
    """simple hello world job to NCBJ"""

    J = baseToAllJobs("helloWorldNCBJ")
    try:
        J.setInputSandbox([find_all("exe-script.py", rootPath, "DIRAC/tests/Workflow")[0]])
    except IndexError:  # we are in Jenkins
        J.setInputSandbox([find_all("exe-script.py", os.environ["WORKSPACE"], "DIRAC/tests/Workflow")[0]])
    J.setExecutable("exe-script.py", "", "helloWorld.log")
    J.setDestination("LCG.NCBJ.pl")
    return endOfAllJobs(J)


def helloWorldGRIDKA():
    """simple hello world job to GRIDKA"""

    J = baseToAllJobs("helloWorldGRIDKA")
    try:
        J.setInputSandbox([find_all("exe-script.py", rootPath, "DIRAC/tests/Workflow")[0]])
    except IndexError:  # we are in Jenkins
        J.setInputSandbox([find_all("exe-script.py", os.environ["WORKSPACE"], "DIRAC/tests/Workflow")[0]])
    J.setExecutable("exe-script.py", "", "helloWorld.log")
    J.setDestination("LCG.GRIDKA.de")
    return endOfAllJobs(J)


def helloWorldGRIF():
    """simple hello world job to GRIF"""

    J = baseToAllJobs("helloWorldGRIF")
    try:
        J.setInputSandbox([find_all("exe-script.py", rootPath, "DIRAC/tests/Workflow")[0]])
    except IndexError:  # we are in Jenkins
        J.setInputSandbox([find_all("exe-script.py", os.environ["WORKSPACE"], "DIRAC/tests/Workflow")[0]])
    J.setExecutable("exe-script.py", "", "helloWorld.log")
    J.setDestination("LCG.GRIF.fr")
    return endOfAllJobs(J)


def helloWorldSSHBatch():
    """simple hello world job to DIRAC.Jenkins_SSHBatch.ch"""

    J = baseToAllJobs("helloWorldSSHBatch")
    try:
        J.setInputSandbox([find_all("exe-script.py", rootPath, "DIRAC/tests/Workflow")[0]])
    except IndexError:  # we are in Jenkins
        J.setInputSandbox([find_all("exe-script.py", os.environ["WORKSPACE"], "DIRAC/tests/Workflow")[0]])
    J.setExecutable("exe-script.py", "", "helloWorld.log")
    J.setDestination("DIRAC.Jenkins_SSHBatch.ch")
    return endOfAllJobs(J)


def helloWorldCloud():
    """simple hello world job to Cloud at Imperial College"""

    J = baseToAllJobs("helloCloudWorld")
    try:
        J.setInputSandbox([find_all("exe-script.py", rootPath, "DIRAC/tests/Workflow")[0]])
    except IndexError:  # we are in Jenkins --- are we ?
        J.setInputSandbox([find_all("exe-script.py", os.environ["WORKSPACE"], "DIRAC/tests/Workflow")[0]])
    J.setExecutable("exe-script.py", "", "helloWorldCloud.log")
    J.setDestination("CLOUD.UKI-LT2-IC-HEP.uk")
    return endOfAllJobs(J)


def mpJob():
    """simple hello world job, with 4Processors and MultiProcessor tags"""

    J = baseToAllJobs("mpJob")
    try:
        J.setInputSandbox([find_all("mpTest.py", rootPath, "DIRAC/tests/Utilities")[0]])
    except IndexError:  # we are in Jenkins
        J.setInputSandbox([find_all("mpTest.py", os.environ["WORKSPACE"], "DIRAC/tests/Utilities")[0]])

    J.setExecutable("mpTest.py")
    J.setTag(["4Processors", "MultiProcessor"])
    return endOfAllJobs(J)


def mp3Job():
    """simple hello world job, with 2 to 4 processors"""

    J = baseToAllJobs("min2max4Job")
    try:
        J.setInputSandbox([find_all("mpTest.py", rootPath, "DIRAC/tests/Utilities")[0]])
    except IndexError:  # we are in Jenkins
        J.setInputSandbox([find_all("mpTest.py", os.environ["WORKSPACE"], "DIRAC/tests/Utilities")[0]])

    J.setExecutable("mpTest.py")
    J.setNumberOfProcessors(numberOfProcessors=3)
    return endOfAllJobs(J)


def min2max4Job():
    """simple hello world job, with 2 to 4 processors"""

    J = baseToAllJobs("min2max4Job")
    try:
        J.setInputSandbox([find_all("mpTest.py", rootPath, "DIRAC/tests/Utilities")[0]])
    except IndexError:  # we are in Jenkins
        J.setInputSandbox([find_all("mpTest.py", os.environ["WORKSPACE"], "DIRAC/tests/Utilities")[0]])

    J.setExecutable("mpTest.py")
    J.setNumberOfProcessors(minNumberOfProcessors=2, maxNumberOfProcessors=4)
    return endOfAllJobs(J)


def wholeNodeJob():
    """simple hello world job, with WholeNode and MultiProcessor tags"""

    J = baseToAllJobs("wholeNodeJob")
    try:
        J.setInputSandbox([find_all("mpTest.py", rootPath, "DIRAC/tests/Utilities")[0]])
    except IndexError:  # we are in Jenkins
        J.setInputSandbox([find_all("mpTest.py", os.environ["WORKSPACE"], "DIRAC/tests/Utilities")[0]])

    J.setExecutable("mpTest.py")
    J.setTag(["WholeNode", "MultiProcessor"])
    return endOfAllJobs(J)


def parametricJob():
    """Creates a parametric job with 3 subjobs which are simple hello world jobs"""

    J = baseToAllJobs("parametricJob")
    try:
        J.setInputSandbox([find_all("exe-script.py", rootPath, "DIRAC/tests/Workflow")[0]])
    except IndexError:  # we are in Jenkins
        J.setInputSandbox([find_all("exe-script.py", os.environ["WORKSPACE"], "DIRAC/tests/Workflow")[0]])
    J.setParameterSequence("args", ["one", "two", "three"])
    J.setParameterSequence("iargs", [1, 2, 3])
    J.setExecutable("exe-script.py", arguments=": testing %(args)s %(iargs)s", logFile="helloWorld_%n.log")
    return endOfAllJobs(J)


def jobWithOutput():
    """Creates a job that uploads an output.
    The output SE is not set here, so it would use the default /Resources/StorageElementGroups/SE-USER
    And possibly use for failover /Resources/StorageElementGroups/Tier1-Failover
    """

    timenow = time.strftime("%s")
    inp1 = [os.path.join(os.getcwd(), timenow + "testFileUpload.txt")]
    with open(inp1[0], "wt") as f:
        f.write(timenow)

    J = baseToAllJobs("jobWithOutput")
    try:
        inp2 = [find_all("exe-script.py", rootPath, "DIRAC/tests/Workflow")[0]]
        J.setInputSandbox(inp1 + inp2)
    except IndexError:  # we are in Jenkins
        inp2 = [find_all("exe-script.py", os.environ["WORKSPACE"], "DIRAC/tests/Workflow")[0]]
        J.setInputSandbox(inp1 + inp2)
    J.setExecutable("exe-script.py", "", "helloWorld.log")
    J.setOutputData([timenow + "testFileUpload.txt"])
    res = endOfAllJobs(J)
    try:
        os.remove(os.path.join(os.getcwd(), timenow + "testFileUpload.txt"))
    except OSError as e:
        return e.errno == errno.ENOENT
    return res


def jobWithOutputs():
    """Creates a job that uploads an output to 2 SEs.
    The output SE is set here to ['RAL-SE', 'IN2P3-SE']
    """

    timenow = time.strftime("%s")
    inp1 = [os.path.join(os.getcwd(), timenow + "testFileUpload.txt")]
    with open(inp1[0], "wt") as f:
        f.write(timenow)

    J = baseToAllJobs("jobWithOutputs")
    try:
        inp2 = [find_all("exe-script.py", rootPath, "DIRAC/tests/Workflow")[0]]
        J.setInputSandbox(inp1 + inp2)
    except IndexError:  # we are in Jenkins
        inp2 = [find_all("exe-script.py", os.environ["WORKSPACE"], "DIRAC/tests/Workflow")[0]]
        J.setInputSandbox(inp1 + inp2)
    J.setExecutable("exe-script.py", "", "helloWorld.log")
    J.setOutputData([timenow + "testFileUpload.txt"], outputSE=["RAL-SE", "IN2P3-SE"])
    res = endOfAllJobs(J)
    try:
        os.remove(os.path.join(os.getcwd(), timenow + "testFileUpload.txt"))
    except OSError as e:
        return e.errno == errno.ENOENT
    return res



########################################################################
# File : Utilities.py
# Author : Federico Stagni
########################################################################

"""
Utilities for Transformation system
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import ast
import random

from DIRAC import S_OK, S_ERROR, gLogger

from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Core.Utilities.SiteSEMapping import getSitesForSE

# from DIRAC.Core.Utilities.Time import timeThis
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

__RCSID__ = "$Id$"


class PluginUtilities(object):
    """
    Utility class used by plugins
    """

    def __init__(self, plugin="Standard", transClient=None, dataManager=None, fc=None, debug=False, transID=None):
        """
        c'tor

        Setting defaults
        """
        # clients
        if transClient is None:
            self.transClient = TransformationClient()
        else:
            self.transClient = transClient
        if dataManager is None:
            self.dm = DataManager()
        else:
            self.dm = dataManager
        if fc is None:
            self.fc = FileCatalog()
        else:
            self.fc = fc

        self.dmsHelper = DMSHelpers()

        self.plugin = plugin
        self.transID = transID
        self.params = {}
        self.groupSize = 0
        self.maxFiles = 0
        self.cachedLFNSize = {}
        self.transString = ""
        self.debug = debug

        self.log = gLogger.getSubLogger(self.plugin + str(self.transID))
        # FIXME: This doesn't work (yet) but should soon, will allow scripts to get the context
        self.log.showHeaders(True)

    def logVerbose(self, message, param=""):
        """logger helper"""
        if self.debug:
            log = gLogger.getSubLogger(self.plugin + " (V)" + str(self.transID))
            log.info(message, param)
        else:
            self.log.verbose(message, param)

    def logDebug(self, message, param=""):
        """logger helper"""
        self.log.debug(message, param)

    def logInfo(self, message, param=""):
        """logger helper"""
        self.log.info(message, param)

    def logWarn(self, message, param=""):
        """logger helper"""
        self.log.warn(message, param)

    def logError(self, message, param=""):
        """logger helper"""
        self.log.error(message, param)

    def logException(self, message, param="", lException=False):
        """logger helper"""
        self.log.exception(message, param, lException)

    def setParameters(self, params):
        """Set the transformation parameters and extract transID"""
        self.params = params
        self.transID = params["TransformationID"]
        self.log = gLogger.getSubLogger(self.plugin + str(self.transID))

    # @timeThis
    def groupByReplicas(self, files, status):
        """
         Generates tasks based on the location of the input data

        :param dict fileReplicas:
                   {'/this/is/at.1': ['SE1'],
                    '/this/is/at.12': ['SE1', 'SE2'],
                    '/this/is/at.2': ['SE2'],
                    '/this/is/at_123': ['SE1', 'SE2', 'SE3'],
                    '/this/is/at_23': ['SE2', 'SE3'],
                    '/this/is/at_4': ['SE4']}

        """
        tasks = []
        nTasks = 0

        if not files:
            return S_OK(tasks)

        files = dict(files)

        # Parameters
        if not self.groupSize:
            self.groupSize = self.getPluginParam("GroupSize", 10)
        flush = status == "Flush"
        self.logVerbose("groupByReplicas: %d files, groupSize %d, flush %s" % (len(files), self.groupSize, flush))

        # Consider files by groups of SEs, a file is only in one group
        # Then consider files site by site, but a file can now be at more than one site
        for groupSE in (True, False):
            if not files:
                break
            seFiles = getFileGroups(files, groupSE=groupSE)
            self.logDebug("fileGroups set: ", seFiles)

            for replicaSE in sortSEs(seFiles):
                lfns = seFiles[replicaSE]
                if lfns:
                    tasksLfns = breakListIntoChunks(lfns, self.groupSize)
                    lfnsInTasks = []
                    for taskLfns in tasksLfns:
                        if flush or (len(taskLfns) >= self.groupSize):
                            tasks.append((replicaSE, taskLfns))
                            lfnsInTasks += taskLfns
                    # In case the file was at more than one site, remove it from the other sites' list
                    # Remove files from global list
                    for lfn in lfnsInTasks:
                        files.pop(lfn)
                    if not groupSE:
                        # Remove files from other SEs
                        for se in [se for se in seFiles if se != replicaSE]:
                            seFiles[se] = [lfn for lfn in seFiles[se] if lfn not in lfnsInTasks]
            self.logVerbose(
                "groupByReplicas: %d tasks created (groupSE %s)" % (len(tasks) - nTasks, str(groupSE)),
                "%d files not included in tasks" % len(files),
            )
            nTasks = len(tasks)

        return S_OK(tasks)

    def createTasksBySize(self, lfns, replicaSE, fileSizes=None, flush=False):
        """
        Split files in groups according to the size and create tasks for a given SE
        """
        tasks = []
        if fileSizes is None:
            fileSizes = self._getFileSize(lfns).get("Value")
        if fileSizes is None:
            self.logWarn("Error getting file sizes, no tasks created")
            return tasks
        taskLfns = []
        taskSize = 0
        if not self.groupSize:
            # input size in GB converted to bytes
            self.groupSize = float(self.getPluginParam("GroupSize", 1.0)) * 1000 * 1000 * 1000
        if not self.maxFiles:
            # FIXME: prepare for chaging the name of the ambiguoug  CS option
            self.maxFiles = self.getPluginParam("MaxFilesPerTask", self.getPluginParam("MaxFiles", 100))
        lfns = sorted(lfns, key=fileSizes.get)
        for lfn in lfns:
            size = fileSizes.get(lfn, 0)
            if size:
                if size > self.groupSize:
                    tasks.append((replicaSE, [lfn]))
                else:
                    taskSize += size
                    taskLfns.append(lfn)
                    if (taskSize > self.groupSize) or (len(taskLfns) >= self.maxFiles):
                        tasks.append((replicaSE, taskLfns))
                        taskLfns = []
                        taskSize = 0
        if flush and taskLfns:
            tasks.append((replicaSE, taskLfns))
        if not tasks and not flush and taskLfns:
            self.logVerbose(
                "Not enough data to create a task, and flush not set (%d bytes for groupSize %d)"
                % (taskSize, self.groupSize)
            )
        return tasks

    # @timeThis
    def groupBySize(self, files, status):
        """
        Generate a task for a given amount of data
        """
        tasks = []
        nTasks = 0

        if not len(files):
            return S_OK(tasks)

        files = dict(files)
        # Parameters
        if not self.groupSize:
            # input size in GB converted to bytes
            self.groupSize = float(self.getPluginParam("GroupSize", 1)) * 1000 * 1000 * 1000
        flush = status == "Flush"
        self.logVerbose("groupBySize: %d files, groupSize: %d, flush: %s" % (len(files), self.groupSize, flush))

        # Get the file sizes
        res = self._getFileSize(list(files))
        if not res["OK"]:
            return res
        fileSizes = res["Value"]

        for groupSE in (True, False):
            if not files:
                break
            seFiles = getFileGroups(files, groupSE=groupSE)

            for replicaSE in sorted(seFiles) if groupSE else sortSEs(seFiles):
                lfns = seFiles[replicaSE]
                newTasks = self.createTasksBySize(lfns, replicaSE, fileSizes=fileSizes, flush=flush)
                lfnsInTasks = []
                for task in newTasks:
                    lfnsInTasks += task[1]
                tasks += newTasks

                # Remove the selected files from the size cache
                self.clearCachedFileSize(lfnsInTasks)
                if not groupSE:
                    # Remove files from other SEs
                    for se in [se for se in seFiles if se != replicaSE]:
                        seFiles[se] = [lfn for lfn in seFiles[se] if lfn not in lfnsInTasks]
                # Remove files from global list
                for lfn in lfnsInTasks:
                    files.pop(lfn)

            self.logVerbose("groupBySize: %d tasks created with groupSE %s" % (len(tasks) - nTasks, str(groupSE)))
            self.logVerbose("groupBySize: %d files have not been included in tasks" % len(files))
            nTasks = len(tasks)

        self.logVerbose("Grouped %d files by size" % len(files))
        return S_OK(tasks)

    def getExistingCounters(self, normalise=False, requestedSites=[]):
        res = self.transClient.getCounters(
            "TransformationFiles", ["UsedSE"], {"TransformationID": self.params["TransformationID"]}
        )
        if not res["OK"]:
            return res
        usageDict = {}
        for usedDict, count in res["Value"]:
            usedSE = usedDict["UsedSE"]
            if usedSE != "Unknown":
                usageDict[usedSE] = count
        if requestedSites:
            siteDict = {}
            for se, count in usageDict.items():
                res = getSitesForSE(se)
                if not res["OK"]:
                    return res
                for site in res["Value"]:
                    if site in requestedSites:
                        siteDict[site] = count
            usageDict = siteDict.copy()
        if normalise:
            usageDict = self._normaliseShares(usageDict)
        return S_OK(usageDict)

    # @timeThis
    def _getFileSize(self, lfns):
        """Get file size from a cache, if not from the catalog
        #FIXME: have to fill the cachedLFNSize!
        """
        lfns = list(lfns)
        cachedLFNSize = dict(self.cachedLFNSize)

        fileSizes = {}
        for lfn in [lfn for lfn in lfns if lfn in cachedLFNSize]:
            fileSizes[lfn] = cachedLFNSize[lfn]
        self.logDebug("Found cache hit for File size for %d files out of %d" % (len(fileSizes), len(lfns)))
        lfns = [lfn for lfn in lfns if lfn not in cachedLFNSize]
        if lfns:
            fileSizes = self._getFileSizeFromCatalog(lfns, fileSizes)
            if not fileSizes["OK"]:
                self.logError(fileSizes["Message"])
                return fileSizes
            fileSizes = fileSizes["Value"]
        return S_OK(fileSizes)

    # @timeThis
    def _getFileSizeFromCatalog(self, lfns, fileSizes):
        """
        Get file size from the catalog
        """
        lfns = list(lfns)
        fileSizes = dict(fileSizes)

        res = self.fc.getFileSize(lfns)
        if not res["OK"]:
            return S_ERROR("Failed to get sizes for all files: %s" % res["Message"])
        if res["Value"]["Failed"]:
            errorReason = sorted(set(res["Value"]["Failed"].values()))
            self.logWarn("Failed to get sizes for %d files:" % len(res["Value"]["Failed"]), errorReason)
        fileSizes.update(res["Value"]["Successful"])
        self.cachedLFNSize.update((res["Value"]["Successful"]))
        self.logVerbose("Got size of %d files from catalog" % len(lfns))
        return S_OK(fileSizes)

    def clearCachedFileSize(self, lfns):
        """Utility function"""
        for lfn in [lfn for lfn in lfns if lfn in self.cachedLFNSize]:
            self.cachedLFNSize.pop(lfn)

    def getPluginParam(self, name, default=None):
        """Get plugin parameters using specific settings or settings defined in the CS
        Caution: the type returned is that of the default value
        """
        # get the value of a parameter looking 1st in the CS
        if default is not None:
            valueType = type(default)
        else:
            valueType = None
        # First look at a generic value...
        optionPath = "TransformationPlugins/%s" % (name)
        value = Operations().getValue(optionPath, None)
        self.logVerbose("Default plugin param %s: '%s'" % (optionPath, value))
        # Then look at a plugin-specific value
        optionPath = "TransformationPlugins/%s/%s" % (self.plugin, name)
        value = Operations().getValue(optionPath, value)
        self.logVerbose("Specific plugin param %s: '%s'" % (optionPath, value))
        if value is not None:
            default = value
        # Finally look at a transformation-specific parameter
        value = self.params.get(name, default)
        self.logVerbose("Transformation plugin param %s: '%s'. Convert to %s" % (name, value, str(valueType)))
        if valueType and not isinstance(value, valueType):
            if valueType is list:
                try:
                    value = ast.literal_eval(value) if value and value != "None" else []
                # literal_eval('SE-DST') -> ValueError
                # literal_eval('SE_MC-DST') -> SyntaxError
                # Don't ask...
                except (ValueError, SyntaxError):
                    value = [val for val in value.replace(" ", "").split(",") if val]

            elif valueType is int:
                value = int(value)
            elif valueType is float:
                value = float(value)
            elif valueType is bool:
                if value in ("False", "No", "None", None, 0):
                    value = False
                else:
                    value = bool(value)
            elif valueType is not str:
                self.logWarn("Unknown parameter type (%s) for %s, passed as string" % (str(valueType), name))
        self.logVerbose("Final plugin param %s: '%s'" % (name, value))
        return value

    @staticmethod
    def _normaliseShares(originalShares):
        """Normalize shares to 1"""
        total = sum(float(share) for share in originalShares.values())
        return dict([(site, 100.0 * float(share) / total if total else 0.0) for site, share in originalShares.items()])

    def uniqueSEs(self, ses):
        """return a list of SEs that are not physically the same"""
        newSEs = []
        for se in ses:
            if not self.isSameSEInList(se, newSEs):
                newSEs.append(se)
        return newSEs

    def isSameSE(self, se1, se2):
        """Check if 2 SEs are indeed the same.

        :param se1: name of the first StorageElement
        :param se2: name of the second StorageElement

        :returns: True/False if they are considered the same.
                  See :py:mod:`~DIRAC.Resources.Storage.StorageElement.StorageElementItem.isSameSE`
        """
        if se1 == se2:
            return True

        return StorageElement(se1).isSameSE(StorageElement(se2))

    def isSameSEInList(self, se1, seList):
        """Check if an SE is the same as any in a list"""
        if se1 in seList:
            return True
        for se in seList:
            if self.isSameSE(se1, se):
                return True
        return False

    def closerSEs(self, existingSEs, targetSEs, local=False):
        """Order the targetSEs such that the first ones are closer to existingSEs. Keep all elements in targetSEs"""
        setTarget = set(targetSEs)
        sameSEs = set([se1 for se1 in setTarget for se2 in existingSEs if self.isSameSE(se1, se2)])
        targetSEs = setTarget - set(sameSEs)
        if targetSEs:
            # Some SEs are left, look for sites
            existingSites = [self.dmsHelper.getLocalSiteForSE(se).get("Value") for se in existingSEs]
            existingSites = set([site for site in existingSites if site])
            closeSEs = set(
                [se for se in targetSEs if self.dmsHelper.getLocalSiteForSE(se).get("Value") in existingSites]
            )
            # print existingSEs, existingSites, targetSEs, closeSEs
            otherSEs = targetSEs - closeSEs
            targetSEs = list(closeSEs)
            random.shuffle(targetSEs)
            if not local and otherSEs:
                otherSEs = list(otherSEs)
                random.shuffle(otherSEs)
                targetSEs += otherSEs
        else:
            targetSEs = []
        return (targetSEs + list(sameSEs)) if not local else targetSEs

    @staticmethod
    def seParamtoList(inputParam):
        """Transform ``inputParam`` to list.

        :param inputParam: can be string, list, or string representation of list
        :returns: list
        """
        if not inputParam:
            return []
        if inputParam.count("["):
            return eval(inputParam)  # pylint: disable=eval-used
        elif isinstance(inputParam, list):
            return inputParam
        return [inputParam]


def getFileGroups(fileReplicas, groupSE=True):
    """
    Group files by set of SEs

    :param dict fileReplicas:
                {'/this/is/at.1': ['SE1'],
                 '/this/is/at.12': ['SE1', 'SE2'],
                 '/this/is/at.2': ['SE2'],
                 '/this/is/at_123': ['SE1', 'SE2', 'SE3'],
                 '/this/is/at_23': ['SE2', 'SE3'],
                 '/this/is/at_4': ['SE4']}

    If groupSE == False, group by SE, in which case a file can be in more than one element
    """
    fileGroups = {}
    for lfn, replicas in fileReplicas.items():
        if not replicas:
            continue
        replicas = sorted(list(set(replicas)))
        if not groupSE or len(replicas) == 1:
            for rep in replicas:
                fileGroups.setdefault(rep, []).append(lfn)
        else:
            replicaSEs = ",".join(replicas)
            fileGroups.setdefault(replicaSEs, []).append(lfn)
    return fileGroups


def sortSEs(ses):
    """Returnes an ordered list of SEs, disk first"""
    seSvcClass = {}
    for se in ses:
        if len(se.split(",")) != 1:
            return sorted(ses)
        if se not in seSvcClass:
            seSvcClass[se] = StorageElement(se).status()["DiskSE"]
    diskSEs = [se for se in ses if seSvcClass[se]]
    tapeSEs = [se for se in ses if se not in diskSEs]
    return sorted(diskSEs) + sorted(tapeSEs)


def sortExistingSEs(lfnSEs, lfns=None):
    """Sort SEs according to the number of files in each (most first)"""
    seFrequency = {}
    archiveSEs = []
    if not lfns:
        lfns = list(lfnSEs)
    else:
        lfns = [lfn for lfn in lfns if lfn in lfnSEs]
    for lfn in lfns:
        existingSEs = lfnSEs[lfn]
        archiveSEs += [s for s in existingSEs if isArchive(s) and s not in archiveSEs]
        for se in [s for s in existingSEs if not isFailover(s) and s not in archiveSEs]:
            seFrequency[se] = seFrequency.setdefault(se, 0) + 1
    sortedSEs = list(seFrequency)
    # sort SEs in reverse order of frequency
    sortedSEs.sort(key=seFrequency.get, reverse=True)
    # add the archive SEs at the end
    return sortedSEs + archiveSEs


def isArchive(se):
    """Is the SE an archive"""
    return DMSHelpers().isSEArchive(se)


def isFailover(se):
    """Is the SE a failover SE"""
    return DMSHelpers().isSEFailover(se)


def getActiveSEs(seList, access="Write"):
    """Utility function - uses the StorageElement cached status"""
    return [se for se in seList if StorageElement(se).status().get(access, False)]
